"""Utility function to create a one figure multi panel validation plot
for continuum imaging of RACS data
"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import NamedTuple

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from astropy.wcs import WCS
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy import stats

from flint.catalogue import Catalogue, get_reference_catalogue
from flint.logging import logger
from flint.naming import processed_ms_format
from flint.summary import BeamSummary, FieldSummary
from flint.utils import estimate_image_centre, get_packaged_resource_path

F_SMALL = 7
F_MED = 8
F_LARGE = 12
F_HUGE = 20
MAX_SOURCES_TO_PLOT = 10000


class ValidationCatalogues(NamedTuple):
    """Container for all the catalogues that are loaded in and
    used throughout validation processing"""

    nvss: Catalogue
    """NVSS catalogue"""
    sumss: Catalogue
    """SUMSS catalogue"""
    icrf: Catalogue
    """ICRF catalogue"""
    askap: Catalogue
    """ASKAP catalogue"""
    racs_high: Catalogue | None = None
    """RACS high catalogue"""
    racs_mid: Catalogue | None = None
    """RACS mid catalogue"""
    tgss: Catalogue | None = None
    """TGSS catalogue"""
    vlass: Catalogue | None = None
    """VLASS catalogue"""


class Tables(NamedTuple):
    """Container for all the tables that are loaded in"""

    nvss: Table
    """NVSS catalogue"""
    sumss: Table
    """SUMSS catalogue"""
    icrf: Table
    """ICRF catalogue"""
    askap: Table
    """ASKAP catalogue"""
    racs_high: Table | None = None
    """RACS high catalogue"""
    racs_mid: Table | None = None
    """RACS mid catalogue"""
    tgss: Table | None = None
    """TGSS catalogue"""
    vlass: Table | None = None
    """VLASS catalogue"""


class XMatchTables(NamedTuple):
    """Container for all the cross matched tables"""

    nvss: Path
    """NVSS catalogue"""
    sumss: Path
    """SUMSS catalogue"""
    icrf: Path
    """ICRF catalogue"""
    racs_high: Path | None = None
    """RACS high catalogue"""
    racs_mid: Path | None = None
    """RACS mid catalogue"""
    tgss: Path | None = None
    """TGSS catalogue"""
    vlass: Path | None = None
    """VLASS catalogue"""


class ValidationTables(NamedTuple):
    """Container for all the tables that are generated by the validation routine"""

    psf_table_path: Path
    """Path to the PSF table"""
    stats_table_path: Path
    """Path to the statistics table"""
    xmatch_tables: XMatchTables
    """Cross-matched tables"""


class ValidatorLayout(NamedTuple):
    """Simple container for all the matplotlib axes objects"""

    ax_rms: Axes
    """Axes for the RMS of the field"""
    ax_legend: Axes
    """Container for basic SBID information"""
    ax_psf: Axes
    """Axes for the PSF of the image"""
    ax_counts: Axes
    """Axes for quick look source counts"""
    ax_brightness1: Axes
    """Axes to compare brightness of sources from the first catalogue"""
    ax_brightness2: Axes
    """Axes to compare brightness of sources from the first catalogue"""
    ax_astrometry: Axes
    """Axes to compare astrometry of sources"""
    ax_astrometry1: Axes
    """Axes to compare astrometry of sources from the first catalogue"""
    ax_astrometry2: Axes
    """Axes to compare astromnetry of sources from the first catalogue"""
    ax_flag_summary: Axes
    """Axes to summarise the flagging of the MS"""


class RMSImageInfo(NamedTuple):
    """Class to hold basic RMS Image information, excluding the actual raw image data"""

    path: Path
    """Path to the RMS fits image"""
    header: fits.Header
    """Header from the FITS image"""
    shape: tuple[int, int]
    """Dimension of the image"""
    no_valid_pixels: int
    """Number of valid pixels in the image"""
    area: float
    """The area of valid sky, in degrees squared"""
    wcs: WCS
    """The WCS of the image"""
    centre: SkyCoord
    """The centre of the image"""
    median: float
    """The median value of the image"""
    minimum: float
    """The minimum value of the image"""
    maximum: float
    """The maximum value of the image"""
    std: float
    """The standard deviation of the image"""
    mode: float
    """The mode of the image"""


class SourceCounts(NamedTuple):
    """A small container to pass around source count information"""

    bins: np.ndarray
    """The bin edges of the source counts in Jy"""
    bin_center: np.ndarray
    """The bin centers to use in Jy"""
    counts_per_bin: np.ndarray
    """The counts of sources per flux bin"""
    counts_per_bin_err: np.ndarray
    """The rough estimate on the errors per bin"""
    euclid_counts: np.ndarray
    """Euclidean normalised source counts"""
    euclid_counts_err: np.ndarray
    """Rough estimate of error on the euclidean normalised source counts"""
    area: float
    """The area in square degrees that the sources cover, i.e. image footprint sky-area"""
    area_fraction: np.ndarray | None = None
    """The fraction of the image that was above a sigma level per flux bin. This may be used as a rough term to scale the Euclidean Normalised source counts. This is not intended to be a robust way of correcting the source counts - just quick"""


class MatchResult(NamedTuple):
    """Simple container to hold results of a quick cross match"""

    name1: str
    """Name of the first survey"""
    name2: str
    """Name of the second survey"""
    pos1: SkyCoord
    """Sky-positions of sources in the first survey"""
    pos2: SkyCoord
    """Sky-positions of sources in the second survey"""
    freq1: float
    """Frequency in Hz of the first survey"""
    freq2: float
    """Frequency in Hz of the second survey"""
    idx1: np.ndarray
    """The indices of the matched sources from the original input table 1"""
    idx2: np.ndarray
    """The indices of the matched sources from the original input table 2"""
    flux1: np.ndarray | None = None
    """Brightness in Jy of source in the first survey"""
    flux2: np.ndarray | None = None
    """Brightness in Jy of source in the second survey"""


def extract_inner_image_array_region(image: np.ndarray, fraction: float) -> np.ndarray:
    """Extract an inner region of an image array. The size of the extracted array
    is specified via the `fraction` parameter, which should be in the range of 0 to 1.

    Args:
        image (np.ndarray): The image to extract pixels from
        fraction (float): The size of the region to extraction.

    Returns:
        np.ndarray: The extracted sub-region
    """
    assert 0.0 < fraction <= 1.0, f"{fraction=}, but has to be between 0.0 to 1.0"

    image_shape = image.shape
    yx_shape = np.array(image_shape[-2:])

    yx_center = np.ceil(yx_shape / 2).astype(int)
    box_size = (yx_shape * fraction).astype(int)
    box_delta = np.ceil(box_size / 2).astype(int)

    min_y = max(0, yx_center[0] - box_delta[0])
    max_y = min(yx_shape[0], yx_center[0] + box_delta[0])

    min_x = max(0, yx_center[1] - box_delta[1])
    max_x = min(yx_shape[1], yx_center[1] + box_delta[1])

    sub_image_array = image[..., min_y:max_y, min_x:max_x]

    return sub_image_array


def get_known_catalogue_info(name: str) -> Catalogue:
    """Return the parameters of a recognised catalogue.

    These are currently hardcoded.

    The structure returneed outlines the name of columns of interest:
    - RA
    - Dec
    - Integrated flux
    - Major / Minor / PA

    Args:
        name (str): The survey name of interest

    Raises:
        ValueError: Raised when an unrecongised catalogue is provided

    Returns:
        Catalogue: Information of the survey catalogue
    """
    # TODO: This should be expanded to include the units as well
    # TODO: Catalogues here need to be packaged somehow

    name = name.upper()

    if name == "NVSS":
        catalogue = Catalogue(
            survey="NVSS",
            file_name="VIII_65_nvss.dat_CH_2.fits",
            name_col="NVSS",
            freq=1.4e9,
            ra_col="RA",
            dec_col="Dec",
            flux_col="S1.4",
            maj_col="MajAxis",
            min_col="MinAxis",
            pa_col="PA",
        )
    elif name == "SUMSS":
        catalogue = Catalogue(
            survey="SUMSS",
            file_name="sumsscat.Mar-11-2008_CLH.fits",
            freq=8.43e8,
            ra_col="RA",
            dec_col="DEC",
            name_col="Mosaic",
            flux_col="St",
            maj_col="dMajAxis",
            min_col="dMinAxis",
            pa_col="dPA",
        )
    elif name == "ICRF":
        catalogue = Catalogue(
            survey="ICRF",
            file_name="icrf.csv",
            freq=1e9,
            ra_col="RA",
            dec_col="Dec",
            name_col="IERS Des.",
            flux_col="None",
            maj_col="None",
            min_col="None",
            pa_col="None",
        )
    else:
        raise ValueError(f"Unknown catalogue {name}")

    return catalogue


def load_known_catalogue(
    name: str, reference_catalogue_directory: Path
) -> tuple[Table, Catalogue]:
    """Load in a known catalogue table

    Args:
        name (str): Name of the survey to load
        reference_catalogue_directory (Path): The directory location with the reference catalogues installed

    Returns:
        Tuple[Table,Catalogue]: The loaded table and Catalogue structure describing the columns
    """
    table, catalogue = get_reference_catalogue(
        reference_directory=reference_catalogue_directory, survey=name
    )
    return table, catalogue


def get_rms_image_info(rms_path: Path, extract_fraction: float = 0.5) -> RMSImageInfo:
    """Extract information about the RMS image and construct a representative structure

    When computing the pixel statistics an inner region of the RMS map is extracted
    to avoid the primary beam roll off. Other properties of the ``RMSImageInfo`` are
    computed on the whole image (e.g. area)

    Args:
        rms_path (Path): The RMS image that will be presented
        extract_fraction (float, optional): Extract the inner region of the base RMS image map. Defaults to 0.5.

    Returns:
        RMSImageInfo: Extracted RMS image information
    """

    with fits.open(rms_path) as rms_image:
        rms_header = rms_image[0].header  # type: ignore
        rms_data = rms_image[0].data  # type: ignore

    logger.info(f"Extracting inner region of {extract_fraction=} for RMS statistics")
    sub_rms_data = extract_inner_image_array_region(
        image=rms_data, fraction=extract_fraction
    )

    valid_pixels = np.sum(np.isfinite(rms_data) & (rms_data != 0))

    ra_cell_size = rms_header["CDELT1"]
    dec_cell_size = rms_header["CDELT2"]
    pixel_area = np.abs(ra_cell_size * dec_cell_size)

    area = pixel_area * valid_pixels

    centre_sky = estimate_image_centre(image_path=rms_path)
    wcs = WCS(fits.getheader(rms_path))

    rms_image_info = RMSImageInfo(
        path=rms_path,
        header=rms_header,
        shape=rms_data.shape,
        no_valid_pixels=valid_pixels,
        area=area,
        wcs=wcs,
        centre=centre_sky,
        median=float(np.nanmedian(sub_rms_data)),
        minimum=float(np.nanmin(sub_rms_data)),
        maximum=float(np.nanmax(sub_rms_data)),
        std=float(np.nanstd(sub_rms_data)),
        mode=stats.mode(sub_rms_data, keepdims=False).mode[0],
    )

    return rms_image_info


def make_validator_axes_layout(fig: Figure, rms_path: Path) -> ValidatorLayout:
    """Create the figure layout to use for the quick look validation plot.

    Args:
        fig (Figure): The figure canvas to add the axes to
        rms_path (Path): Path to the RMS image that will be presented. Loaded to access the WCS

    Returns:
        ValidatorLayout: Representation of all axes objects
    """
    # My friendship with gridspec is over, now subplot_mosaic is my best friend
    # Using the following encoding:
    # T = text / field info
    # F = flux / source counts
    # A = astrometry
    # P = PSF
    # B = BANE noise map
    # S = SUMMS flux comparison
    # s = SUMMS astrometry comparison
    # N = NVSS flux comparison
    # n = NVSS astrometry comparison
    # f = Flagging summary
    rms_wcs = WCS(fits.getheader(rms_path)).celestial
    ax_dict = fig.subplot_mosaic(
        """
        TFPA
        BBSs
        BBNn
        ffff
        """,
        per_subplot_kw={  # type: ignore
            "B": {"projection": rms_wcs},
        },
        subplot_kw={
            "aspect": "equal",
        },
    )
    for spine in ax_dict["T"].spines.values():
        spine.set_edgecolor("none")
    _ = ax_dict["T"].axes.yaxis.set_visible(False)
    _ = ax_dict["T"].axes.xaxis.set_visible(False)

    # Set the axes that are shared
    _ = ax_dict["N"].sharex(ax_dict["S"])
    _ = ax_dict["n"].sharex(ax_dict["s"])

    validator_layout = ValidatorLayout(
        ax_rms=ax_dict["B"],
        ax_legend=ax_dict["T"],
        ax_psf=ax_dict["P"],
        ax_counts=ax_dict["F"],
        ax_brightness1=ax_dict["N"],
        ax_brightness2=ax_dict["S"],
        ax_astrometry=ax_dict["A"],
        ax_astrometry1=ax_dict["n"],
        ax_astrometry2=ax_dict["s"],
        ax_flag_summary=ax_dict["f"],
    )

    return validator_layout


def plot_flag_summary(
    fig: Figure,
    ax: Axes,
    field_summary: FieldSummary,
) -> Axes:
    """Plot the percentage of the spectrum that is flagged for
    each measurement set

    Args:
        fig (Figure): The figure canvas that is being plotted to
        ax (Axes): The axes object that is being plotted to
        field_summary (FieldSummary): An active field summary object with the collection of MSSummary structures

    Returns:
        Axes: The axes object with the plotted RMS image
    """
    if field_summary.ms_summaries is None:
        logger.info("No measurement summaries to plot")
        ax.text(
            0.5,
            0.5,
            "No data to plot",
            va="center",
            ha="center",
            transform=ax.transAxes,
        )

        return ax

    ms_summaries = sorted(
        [mss for mss in field_summary.ms_summaries], key=lambda x: x.beam
    )
    accumulated_flags = 0
    accumulated_samples = 0
    for ms_summary in ms_summaries:
        flag_spectrum = ms_summary.flag_spectrum
        accumulated_flags += np.sum(np.nan_to_num(flag_spectrum))
        accumulated_samples += len(flag_spectrum)
        ax.plot(
            np.arange(len(flag_spectrum)),
            flag_spectrum,
            label=f"{ms_summary.beam:02d}",
            lw=0.5,
        )

    percent_sbid_flagged = accumulated_flags / accumulated_samples * 100.0
    ax.legend(ncols=18, title="Beam Number", loc="lower left")
    ax.grid()
    ax.axhline(0.0, color="black")
    ax.set(
        xlabel="Channel",
        ylabel="Flagged fraction",
        ylim=(-0.35, 1.1),
        aspect="auto",
        title=f"Flagging summary - {percent_sbid_flagged:.2f}% of {field_summary.sbid} flagged",
    )

    return ax


def plot_rms_map(
    fig: Figure, ax: Axes, rms_path: Path, source_positions: SkyCoord | None = None
) -> Axes:
    """Add the RMS image to the figure

    Args:
        fig (Figure): Figure that contains the axes object
        ax (Axes): The axes that will be plotted
        rms_path (Path): Location of the RMS image
        source_positions (Optional[SkyCoord], optionals): Sources in the ASKAP catalogue to overlay onto the RMS image. Defaults to None.

    Returns:
        Axes: The axes object with the plotted RMS image
    """

    # Convert Jy/beam to uJy/beam
    rms_data = fits.getdata(rms_path) * 1e6  # type: ignore

    # Pirate no believes get below 10uJy/beam mateeey. Percentile a little more
    # robust to outliers but much of a much
    floor: float = max(1, np.floor(np.log10(np.nanpercentile(rms_data, 16))))  # type: ignore

    im = ax.imshow(
        np.log10(rms_data),
        vmin=floor,
        vmax=floor + 1.0,
        origin="lower",
        cmap="YlOrRd",
    )

    if source_positions is not None:
        # If there are too many sources on the RMS map nothing can be seen
        if len(source_positions) > MAX_SOURCES_TO_PLOT:
            logger.info(
                f"{len(source_positions)} sources, too many to plot. Skipping. "
            )
        else:
            ax.scatter(
                source_positions.ra,  # type: ignore
                source_positions.dec,  # type: ignore
                marker="o",
                edgecolor="black",
                facecolor="none",
                s=0.1,
                transform=ax.get_transform("fk5"),  # type: ignore
            )

    ax.grid(color="0.5", ls="solid")
    ax.set_xlabel("RA (J2000)")
    ax.set_ylabel("Dec (J2000)")

    # These offsets are pretty and correspond to 10, 20, 30,m 50, 100 where floor is 10 uJy/bea,
    yt = floor + np.array([0.0, 0.30103, 0.47712125, 0.69897, 1.0])
    ytl = [f"{10**lx:.0f}" for lx in yt]

    cbar = fig.colorbar(
        im,
        ax=ax,
        use_gridspec=True,
        pad=0.09,
        shrink=0.5,
        orientation="vertical",
    )
    cbar.set_label(r"Image rms ($\mu$Jy/beam)", fontsize=F_SMALL)
    cbar.set_ticks(yt)
    cbar.set_ticklabels(ytl)

    gal_col = "b"
    overlay = ax.get_coords_overlay("galactic")
    overlay.grid(color=gal_col, ls="dashed", alpha=0.6)
    overlay[0].tick_params(colors=gal_col)
    overlay[1].tick_params(colors=gal_col)
    overlay[0].set_axislabel("Galactic long.", color=gal_col)
    overlay[1].set_axislabel("Galactic lat.", color=gal_col)

    ax.set_title("Image noise", loc="left", fontsize=F_MED)

    return ax


def calculate_area_correction_per_flux(
    rms_image_path: Path, flux_bin_centre: np.ndarray, sigma: float = 5
) -> np.ndarray:
    """This derives a rough correction to the area term when calculating
    the source counts. This is not intended to correct for completeness,
    although they are closely related.

    The RMS image is read in and a CDF is calculated. This is used to
    return a scaling term that indicates what fraction a flux-bin was
    `sigma` times the RMS, which could be used to scale the area.

    Args:
        rms_image_path (Path): RMS image that will be use to calculate the area correction
        flux_bin_centre (np.ndarray): Fluxes to calculate the area fraction at
        sigma (float, optional): Mutliplicate term indicating what sigma level source finding cropped at. Defaults to 5.

    Returns:
        np.ndarray: Fraction of the image that was available for source finding at a flux density
    """

    # Filter out the data
    rms_data = fits.getdata(rms_image_path)
    assert rms_data is not None, f"{rms_data=}, which should not happen"
    rms_data = rms_data[np.isfinite(rms_data)].flatten()

    # Calculate the cumulative distribution and normalise it to 0..1
    rms_hist, rms_bins = np.histogram(rms_data, bins=500)
    rms_cdf = np.cumsum(rms_hist)
    rms_cdf = rms_cdf / rms_cdf[-1]
    assert np.isclose(rms_cdf[-1], 1), "Normalisation of the RMS CDF failed"

    # Calculate the bin centers of the histogram / CDF
    rms_bin_centre = 0.5 * (rms_bins[1:] + rms_bins[:-1])

    # Now calculate the area available per flux bin that is available
    # for source finding. Since the source finding is often performed
    # with some flux / rms cut, the area available is a multiple of
    # the rms. The scaling term here is a generic value that most
    # pirates use.
    area_fraction_per_bin = np.interp(
        flux_bin_centre, rms_bin_centre * sigma, rms_cdf, left=0.0, right=1.0
    )

    return area_fraction_per_bin


def get_source_counts(
    fluxes: np.ndarray,
    area: float,
    minlogf: float = -4,
    maxlogf: float = 2,
    Nbins: int = 81,
    rms_image_path: Path | None = None,
) -> SourceCounts:
    """Derive source counts for a set of fluxes and known area

    Args:
        fluxes (np.ndarray): The fluxes in Jy to count
        area (float): Area over which the sources were collected
        minlogf (float, optional): The minimum bin edge, in Jy. Defaults to -4.
        maxlogf (float, optional): The maximum bin edgem, in Jy. Defaults to 2.
        Nbins (int, optional): Number of bins to include in the source counts. Defaults to 81.
        rms_image_path (Optional[Path], optional): Path to the RMS image. If not None, it is used to calculate a rough area correction. Defaults to None.

    Returns:
        SourceCounts: Source counts and their properties
    """
    logger.info(
        f"Computing source counts for {len(fluxes)} sources, {Nbins=}, {area=:.2f} sq.deg."
    )

    logf = np.linspace(minlogf, maxlogf, Nbins)
    f = np.power(10.0, logf)

    counts, bins = np.histogram(fluxes, f)
    binsmid = 0.5 * ((bins[:-1]) + (bins[1:]))
    ds = bins[1:] - bins[:-1]

    area_sky_total = 360.0 * 360.0 / np.pi
    solid_angle = 4 * np.pi * area / area_sky_total

    counts_err = np.array(np.sqrt(counts), dtype=int)

    scount = np.power(binsmid, 2.5) * counts / (ds * solid_angle)
    scount_err = np.power(binsmid, 2.5) * 1.0 * counts_err / (ds * solid_angle)

    area_fraction = (
        calculate_area_correction_per_flux(
            rms_image_path=rms_image_path, flux_bin_centre=binsmid, sigma=5
        )
        if rms_image_path
        else None
    )

    source_counts = SourceCounts(
        bins=bins,
        bin_center=0.5 * (bins[1:] + bins[:-1]),
        counts_per_bin=counts,
        counts_per_bin_err=counts_err,
        euclid_counts=scount,
        euclid_counts_err=scount_err,
        area=area,
        area_fraction=area_fraction,
    )

    return source_counts


def plot_source_counts(
    catalogue: Table,
    rms_info: RMSImageInfo,
    ax: Axes,
    freq: float | None = None,
    dezotti: Table | None = None,
    skads: Table | None = None,
) -> Axes:
    """Create a figure of source counts from a astropy Table. If
    `freq` and either `dezotti` / `skads` are supplied then these
    precomputed source count tables are also included in the
    panel.

    When computing the source counts for `catalogue`, only a
    minimumal set of corrections are derived and applied.

    Args:
        catalogue (Table): The catalogue to derive source counts for
        rms_info (RMSImageInfo): Look up information from the RMS file that catalogue was constructed against
        ax (Axes): The axes panel the counts will be plottedd on
        freq (Optional[float], optional): Frequency that the source catalogue. Used to scale the Dezotti and SKADS tables. Defaults to None.
        dezotti (Optional[Table], optional): Loaded reference table of Dezotti source counts. Defaults to None.
        skads (Optional[Table], optional): Loaded reference table of SKADS source counts. Defaults to None.

    Returns:
        Axes: The axes object used for plotting
    """
    # TODO: Is the freq column needed anymore

    fluxes = catalogue["int_flux"]

    source_counts = get_source_counts(
        fluxes=fluxes, area=rms_info.area, rms_image_path=rms_info.path
    )

    ax.errorbar(
        source_counts.bin_center * 1e3,
        source_counts.euclid_counts,
        yerr=source_counts.euclid_counts_err,
        fmt=".",
        color="darkred",
        label="Raw Component Catalogue",
    )

    # Since area_fraction could be a numpy array we have to
    # explicitly test to ensure it is not None, otherwise it
    # is ambiguous
    if source_counts.area_fraction is not None:
        ax.errorbar(
            source_counts.bin_center * 1e3,
            source_counts.euclid_counts / source_counts.area_fraction,
            yerr=source_counts.euclid_counts_err / source_counts.area_fraction,
            fmt="x",
            color="darkred",
            label="Area Corrected Catalogue",
        )

    if dezotti is not None and freq is not None:
        spectral_index_scale = (freq / 1.4e9) ** -0.8
        ax.errorbar(
            np.power(10.0, dezotti["col1"]) * 1.0e3 * spectral_index_scale,
            dezotti["col2"] * np.power(spectral_index_scale, 1.5),
            yerr=[
                dezotti["col4"] * np.power(spectral_index_scale, 1.5),
                dezotti["col3"] * np.power(spectral_index_scale, 1.5),
            ],
            fmt="d",
            color="grey",
            alpha=0.2,
            label="de Zotti et al. 2010",
        )

    if skads is not None and freq is not None:
        spectral_index_scale = (freq / 1.4e9) ** -0.8
        ax.errorbar(
            skads["fl_bin_mid"] * 1.0e3 * spectral_index_scale,
            skads["SC"] * np.power(spectral_index_scale, 1.5),
            yerr=skads["SC_err"] * np.power(spectral_index_scale, 1.5),
            fmt=":",
            color="k",
            label="Wilman et al. 2008",
        )

    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_xlabel("Flux Density (mJy)")
    ax.set_ylabel(r"$\frac{dN}{dS} S^{\frac{5}{2}}$ (Jy$^{\frac{3}{2}}$ sr$^{-1}$)")
    ax.set_title(f"{len(fluxes)} sources")
    ax.set_xlim(0.2, 1.0e4)
    ax.set_ylim(5.0e-2, 1.0e4)
    ax.grid()
    ax.legend(loc="lower right", fontsize=F_SMALL)

    return ax


def match_nearest_neighbour(
    table1: Table,
    table2: Table,
    catalogue1: Catalogue,
    catalogue2: Catalogue,
    radius: float = 10,
) -> MatchResult:
    """Match two catalogues together, and construct common properties.

    Args:
        table1 (Table): The catalogue table from survey one
        table2 (Table): The catalogue table from survey two
        catalogue1 (Catalogue): Catalogue metadata for survey one
        catalogue2 (Catalogue): Catalogue metadata for survey two
        radius (float, optional): Maximum matching radius. Defaults to 10.

    Returns:
        MatchResult: Object containing source matches and common properties
    """

    pos1 = SkyCoord(
        table1[catalogue1.ra_col], table1[catalogue1.dec_col], unit="deg,deg"
    )
    pos2 = SkyCoord(
        table2[catalogue2.ra_col], table2[catalogue2.dec_col], unit="deg,deg"
    )

    idx, sep, _ = pos1.match_to_catalog_sky(catalogcoord=pos2, nthneighbor=1)
    mask = sep < (radius * u.arcsecond)

    idx = idx[mask]

    match_result = MatchResult(
        name1=catalogue1.survey,
        name2=catalogue2.survey,
        pos1=pos1[mask],
        pos2=pos2[idx],
        freq1=catalogue1.freq,
        freq2=catalogue2.freq,
        idx1=np.argwhere(mask)[:, 0],
        idx2=idx,
        flux1=(
            table1[catalogue1.flux_col].value[mask]
            if catalogue1.flux_col != "None"
            else None
        ),
        flux2=(
            table2[catalogue2.flux_col].value[idx]
            if catalogue2.flux_col != "None"
            else None
        ),
    )

    return match_result


def plot_astrometry_comparison(
    fig: Figure, ax: Axes, match_result: MatchResult
) -> Axes:
    """Plot the astrometry of cross matches from a match result set

    Args:
        fig (Figure): The figure canvas plotting on
        ax (Axes): The Axes being plotted on
        match_result (MatchResult): The set of sources cross-matched and found in common

    Returns:
        Axes: The Axes plotted on
    """
    logger.info(
        f"Plotting astrometry match between {match_result.name1} and {match_result.name2}"
    )
    if len(match_result.pos1) == 0:
        ax.set_xlim(-8.0, 8.0)
        ax.set_ylim(-8.0, 8.0)
        ax.text(0.0, 0.0, f"No data for {match_result.name2}", va="center", ha="center")

        return ax

    pos1, pos2 = match_result.pos1, match_result.pos2

    # Get the separations between two points, and their angle between them
    seps = pos1.separation(pos2)
    pas = pos1.position_angle(pos2)

    # Compute the detlaRA and deltaDec. Straight subtraction is not correct
    err_x = (seps * -np.sin(pas.radian)).to(u.arcsec).value
    err_y = (seps * np.cos(pas.radian)).to(u.arcsec).value

    mean_x, mean_y = err_x.mean(), err_y.mean()
    std_x, std_y = err_x.std(), err_y.std()

    ax.plot(err_x, err_y, ".r", zorder=0, ms=2)
    ax.errorbar(mean_x, mean_y, xerr=std_x, yerr=std_y, fmt="ok")

    ax.grid()
    ax.axvline(0, color="black", ls=":")
    ax.axhline(0, color="black", ls=":")

    ax.text(
        -7.0,
        7.0,
        match_result.name2,
        fontsize=F_MED,
        bbox=dict(facecolor="white", edgecolor="none", alpha=0.8),
    )
    ax.text(
        -7.0,
        7.0,
        f"{match_result.name2} - {len(err_x)} matches",
        fontsize=F_MED,
        bbox=dict(facecolor="white", edgecolor="none", alpha=0.8),
    )

    ax.set(
        xlim=(-8, 8), ylim=(-8, 8), xlabel="Offset (arcsec)", ylabel="Offset (arcsec)"
    )
    le_a1 = rf" $\epsilon_{{SU}} : ({{{mean_x:.1f}}}\pm{{{std_x:.1f}}},{{{mean_y:.1f}}}\pm{{{std_y:.1f}}})$"
    ax.text(
        7.6,
        7.6,
        le_a1,
        va="top",
        ha="right",
        fontsize=F_MED,
        bbox=dict(facecolor="white", edgecolor="none", alpha=0.8),
    )
    return ax


def plot_flux_comparison(fig: Figure, ax: Axes, match_result: MatchResult) -> Axes:
    """Create a flux comparison plot showing the flux densities from two catalogues compared
    to one another.

    Args:
        fig (Figure): The figure canvas that the axes is on
        ax (Axes): The axes object that will be render the plot
        match_result (MatchResult): A result set of the cross-match between two catalogues

    Returns:
        Axes: The aces object that was used for plotting
    """
    if len(match_result.pos1) == 0:
        ax.loglog([2.0, 2000.0], [2.0, 2000.0], "ow")
        ax.text(1e0, 1e2, f"No data for {match_result.name2}", va="center", ha="center")

        return ax

    flux1 = match_result.flux1
    flux2 = match_result.flux2

    one2one = np.array([0.002, 8])
    spectral_index_scale = (match_result.freq2 / match_result.freq1) ** -0.8

    ax.loglog(flux1, flux2, "ok", ms=2)
    ax.loglog(one2one, one2one, "-", c="r", label="Raw")
    ax.loglog(one2one, one2one * spectral_index_scale, "--", c="r", label="Scaled")
    ax.set(
        ylabel=f"{match_result.name2} Integrated Flux (Jy)",
        xlabel=f"{match_result.name1} Integrated Flux (Jy)",
    )
    ax.legend()

    ax.grid()

    return ax


def scale_flux_alpha(
    flux: float | np.ndarray,
    freq: float | np.ndarray,
    ref_freq: float,
    alpha: float = -0.8,
) -> float | np.ndarray:
    """Scale a flux density to a reference frequency using a spectral index

    Args:
        flux (float): The flux density to scale
        freq (float): The frequency of the flux density
        ref_freq (float): The reference frequency to scale to
        alpha (float, optional): The spectral index to use. Defaults to -0.8.

    Returns:
        float: The scaled flux density
    """
    return flux * (freq / ref_freq) ** alpha


def make_xmatch_table(
    table1: Table,
    table2: Table,
    catalogue1: Catalogue,
    catalogue2: Catalogue,
    match_result: MatchResult,
    output_path: Path,
) -> tuple[Table, Path]:
    """Create a simple cross match table between two catalogues

    Args:
        table1 (Table): The catalogue table from survey one
        table2 (Table): The catalogue table from survey two
        catalogue1 (Catalogue): Catalogue metadata for survey one
        catalogue2 (Catalogue): Catalogue metadata for survey two
        match_result (MatchResult): A result set of the cross-match between two catalogues
        output_path (Path): Location to save the table to

    Returns:
        Tuple[Table, Path]: The table and the output path
    """
    askap_flux = match_result.flux1
    # Some catalogues (like the ICFS) do not have a flux column
    external_flux = (
        match_result.flux2
        if match_result.flux2 is not None
        else np.ones_like(match_result.idx2) * -1
    )

    askap_flux_scaled = scale_flux_alpha(
        flux=askap_flux,
        freq=match_result.freq1,
        ref_freq=match_result.freq2,
    )

    xmatch_table = Table(
        {
            "Component_ID": table1[catalogue1.name_col][match_result.idx1],
            "RA": table1[catalogue1.ra_col][match_result.idx1],
            "DEC": table1[catalogue1.dec_col][match_result.idx1],
            "ASKAP_flux": table1[catalogue1.flux_col][match_result.idx1],
            "ASKAP_flux_SNR": table1[catalogue1.flux_col][match_result.idx1]
            / table1["local_rms"][match_result.idx1],
            "RA_comp": table2[catalogue2.ra_col][match_result.idx2],
            "DEC_comp": table2[catalogue2.dec_col][match_result.idx2],
            "Comp_flux": external_flux,
            "Flux_Ratio_AlphaCorr": askap_flux_scaled / external_flux,
            "Flux_Ratio_NOAlphaCorr": askap_flux / external_flux,
        }
    )

    suffix = f"{catalogue2.survey}_{catalogue1.survey}"
    output_file = output_path / f"cat_match_{suffix}.csv"
    xmatch_table.write(output_file, format="csv", overwrite=True)

    return xmatch_table, output_file


def plot_psf(fig: Figure, ax: Axes, rms_info: RMSImageInfo) -> Axes:
    """Create a plot highlighting the synthesised beam recorded in the
    RMS image header

    Args:
        fig (Figure): Fogire canvas being used
        ax (Axes): The axes object that will be used for plotting
        rms_info (RMSImageInfo): Extracted information from the RMS image

    Returns:
        Axes: The aces object used for plotting
    """

    bmaj = rms_info.header["BMAJ"] * 3600
    bmin = rms_info.header["BMIN"] * 3600

    ax.plot(bmin, bmaj, "ok", ms=3)
    ax.set(
        xlim=(5.0, 20.0),
        ylim=(5.0, 40.0),
        xlabel="PSF minor axis (arcsec)",
        ylabel="PSF major axis (arcsec)",
    )

    ax.grid()

    return ax


def plot_field_info(
    fig: Figure,
    ax: Axes,
    field_summary: FieldSummary,
    rms_info: RMSImageInfo,
    askap_table: Table,
) -> Axes:
    # TODO: Add the field information to the plot
    # Namely:
    # - Field name
    # - J2000 RA / Dec
    # - Galactic l / b
    # - SBID
    # - CAL_SBID
    # - Start time
    # - Integration time
    # - Hour angle range
    # - Elevation range
    # - Median rms
    # - Number of sources
    # - Processing date

    hour_angles = field_summary.hour_angles
    elevations = field_summary.elevations
    pol_axis_str = (
        f"{field_summary.pol_axis:.3f} rad" if field_summary.pol_axis else "Unknown"
    )
    field_text = "\n".join(
        (
            f"Field name: {field_summary.field_name}",
            f"- J2000 RA / Dec    : {rms_info.centre.icrs.to_string(style='hmsdms', precision=1)}",  # type: ignore
            f"- Galactic l / b    : {rms_info.centre.galactic.to_string(style='decimal')}",  # type: ignore
            f"- SBID              : {field_summary.sbid}",
            f"- CAL_SBID          : {field_summary.cal_sbid}",
            f"- Holorgraphy file  : {field_summary.holography_path.name if field_summary.holography_path else 'N/A'}",
            f"- Start time        : {field_summary.ms_times[0].utc.fits}",  # type: ignore
            f"- Integration time  : {field_summary.integration_time * u.second:latex_inline}",  # type: ignore
            f"- Hour angle range  : {hour_angles.min().to_string(precision=2, format='latex_inline')} - {hour_angles.max().to_string(precision=2, format='latex_inline')}",  # type: ignore
            f"- Elevation range   : {elevations.min().to_string(precision=2, format='latex_inline')} - {elevations.max().to_string(precision=2, format='latex_inline')}",  # type: ignore
            f"- Median rms uJy    : {rms_info.median * 1e6:.1f}",
            f"- Components        : {len(askap_table)}",
            f"- Processing date   : {Time.now().fits}",
            f"- Pol. axis         : {pol_axis_str}",
        )
    )
    props = dict(boxstyle="square", facecolor="none", edgecolor="tab:red", pad=2)
    ax.text(
        0.0,
        0.0,
        field_text,
        family="monospace",
        fontdict={"fontsize": F_LARGE},
        wrap=True,
        bbox=props,
        ha="left",
    )
    return ax


def load_catalogues(
    source_catalogue_path: Path,
    reference_catalogue_directory: Path,
    askap_survey_name: str,
    rms_info: RMSImageInfo,
) -> tuple[ValidationCatalogues, Tables]:
    """Load in all the catalogues that are required for the validation.

    Args:
        source_catalogue_path (Path): The source catalogue to load
        reference_catalogue_directory (Path): The directory that contains the reference ICRF, NVSS and SUMSS catalogues.
        askap_survey_name (str): The name that will be given to the ASKAP field data
        rms_info (RMSImageInfo): The extracted information from the RMS image

    Returns:
        Tuple[ValidationCatalogues, Tables]: The loaded catalogues and tables
    """
    logger.info(f"Loading {source_catalogue_path=}")
    askap_table = Table.read(source_catalogue_path)
    askap_cata = Catalogue(
        survey=askap_survey_name,
        file_name=source_catalogue_path.name,
        freq=rms_info.header["CRVAL3"],
        ra_col="ra",
        dec_col="dec",
        name_col="source",
        flux_col="int_flux",
        maj_col="a",
        min_col="b",
        pa_col="pa",
    )
    icrf_table, icrf_catalogue = load_known_catalogue(
        name="ICRF", reference_catalogue_directory=reference_catalogue_directory
    )
    sumss_table, sumss_catalogue = load_known_catalogue(
        name="SUMSS", reference_catalogue_directory=reference_catalogue_directory
    )
    nvss_table, nvss_catalogue = load_known_catalogue(
        name="NVSS", reference_catalogue_directory=reference_catalogue_directory
    )

    return (
        ValidationCatalogues(
            askap=askap_cata,
            icrf=icrf_catalogue,
            sumss=sumss_catalogue,
            nvss=nvss_catalogue,
        ),
        Tables(
            askap=askap_table,
            icrf=icrf_table,
            sumss=sumss_table,
            nvss=nvss_table,
        ),
    )


def make_field_stats_table(
    field_summary: FieldSummary,
    rms_info: RMSImageInfo,
    output_path: Path,
) -> Path:
    # Columns are:
    # FLD_NAME,SBID,NPIXV,MINIMUM,MAXIMUM,MED_RMS_uJy,MODE_RMS_uJy,STD_RMS_uJy,MIN_RMS_uJy,MAX_RMS_uJy
    # TODO: Get the min / max from the image (not rms)
    if field_summary.linmos_image is not None:
        with fits.open(field_summary.linmos_image) as linmos_image:
            image_data = linmos_image[0].data
            min_image_data = np.nanmin(image_data)
            max_image_data = np.nanmax(image_data)
    else:
        logger.warning("The linmos image in the provided field summary is empty.")
        min_image_data = np.nan
        max_image_data = np.nan

    stats_table = Table(
        {
            "FLD_NAME": [field_summary.field_name],
            "SBID": [field_summary.sbid],
            "NPIXV": [rms_info.no_valid_pixels],
            "MINIMUM": [min_image_data],
            "MAXIMUM": [max_image_data],
            "MED_RMS_uJy": [rms_info.median * 1e6],
            "MODE_RMS_uJy": [rms_info.mode * 1e6],
            "STD_RMS_uJy": [rms_info.std * 1e6],
            "MIN_RMS_uJy": [rms_info.minimum * 1e6],
            "MAX_RMS_uJy": [rms_info.maximum * 1e6],
        }
    )
    # statistics_{SBID}-{FIELD_NAME}.csv
    outfile = (
        output_path / f"statistics_{field_summary.sbid}-{field_summary.field_name}.csv"
    )
    stats_table.write(outfile, format="csv", overwrite=True)

    return outfile


class PSFTableRow(NamedTuple):
    """Something that is only used to internally
    represent the PSF information of a beam image.
    Currently no reason to use this.
    """

    beam: int
    vis_total: int
    vis_flagged: int
    image_name: str
    bmaj: float
    bmin: float
    bpa: float
    ra: float
    dec: float
    l: float  # noqa: E741
    b: float


def _make_beam_psf_row(beam_summary: BeamSummary) -> PSFTableRow:
    """Collects the information required of a single beam image
    and measurement set for entry into the PSF table.

    Not intended for usage other than int he creation of the PSF table.

    Args:
        beam_summary (BeamSummary): The input set of collected properties

    Returns:
        PSFTableRow: Extracted information
    """

    name_components = processed_ms_format(in_name=beam_summary.ms_summary.path)
    vis_total = beam_summary.ms_summary.flagged + beam_summary.ms_summary.unflagged
    vis_flagged = beam_summary.ms_summary.flagged

    assert beam_summary.image_set is not None, (
        f"{beam_summary.image_set=}, which should not happen"
    )
    image_file = list(beam_summary.image_set.image)[-1]
    with fits.open(image_file) as image:
        bmaj = image[0].header["BMAJ"]  # type: ignore
        bmin = image[0].header["BMIN"]  # type: ignore
        bpa = image[0].header["BPA"]  # type: ignore

    coord = estimate_image_centre(image_path=image_file)

    assert name_components is not None, f"{name_components=}, which should not happen"
    return PSFTableRow(
        beam=int(name_components.beam) if name_components.beam else -999,
        vis_flagged=vis_flagged,
        vis_total=vis_total,
        image_name=image_file.name,
        bmaj=bmaj,
        bmin=bmin,
        bpa=bpa,
        ra=coord.ra.deg,
        dec=coord.dec.deg,
        l=coord.galactic.l.deg,
        b=coord.galactic.b.deg,
    )


def make_psf_table(field_summary: FieldSummary, output_path: Path) -> Path:
    # TODO: This will likely need changes to the
    # FieldSummary structure to handle holding MSSummary objects
    # Columns are:
    # BEAM_NUM,BEAM_TIME,RA_DEG,DEC_DEG,GAL_LONG,GAL_LAT,PSF_MAJOR,PSF_MINOR,PSF_ANGLE,VIS_TOTAL,VIS_FLAGGED

    assert field_summary.beam_summaries is not None, (
        f"{field_summary.beam_summaries=}, which should not happen"
    )

    psf_table_rows = [
        _make_beam_psf_row(beam_summary=beam_summary)
        for beam_summary in field_summary.beam_summaries
    ]

    psf_table = Table(
        {
            "BEAM_NUM": [row.beam for row in psf_table_rows],
            "VIS_TOTAL": [row.vis_total for row in psf_table_rows],
            "VIS_FLAGGED": [row.vis_total for row in psf_table_rows],
            "IMAGE_NAME": [row.image_name for row in psf_table_rows],
            "PSF_MAJOR": [row.bmaj for row in psf_table_rows],
            "PSF_MINOR": [row.bmin for row in psf_table_rows],
            "PSF_ANGLE": [row.bpa for row in psf_table_rows],
        }
    )

    psf_table.sort("BEAM_NUM")
    # beam_inf_{SBID}-{FIELD_NAME}.csv
    outfile = (
        output_path / f"beam_inf_{field_summary.sbid}-{field_summary.field_name}.csv"
    )
    psf_table.write(outfile, format="csv", overwrite=True)

    return outfile


def create_validation_tables(
    field_summary: FieldSummary,
    rms_image_path: Path,
    source_catalogue_path: Path,
    output_path: Path,
    reference_catalogue_directory: Path,
) -> ValidationTables:
    """Create a set of validation tables that can be used to assess the
    correctness of an image and associated source catalogue.

    Args:
        field_summary (FieldSummary): A description of the key properties of the field
        rms_image_path (Path): The RMS fits image the source catalogue was constructed against.
        source_catalogue_path (Path): The source catalogue.
        output_path (Path): The output path of the figure to create
        reference_catalogue_directory (Path): The directory that contains the reference catalogues installed

    Returns:
        ValidationTables: The tables that were created
    """
    logger.info("Creating validation tables")

    rms_info = get_rms_image_info(rms_path=rms_image_path)

    askap_survey_name = f"{field_summary.sbid}-{field_summary.field_name}"

    # Load in the catalogues
    catalogues, tables = load_catalogues(
        source_catalogue_path=source_catalogue_path,
        reference_catalogue_directory=reference_catalogue_directory,
        askap_survey_name=askap_survey_name,
        rms_info=rms_info,
    )

    # Loop over all the catalogues and cross match them to the ASKAP catalogue
    _tables: dict[str, Table] = {}
    for survey in catalogues._fields:
        if survey == "askap":
            continue

        # The index method of a named tuple is the same as tuple[i], where
        # i is an int. This is a round about way of getting the key name
        # from a named attributed stored in a parameter
        survey_catalogue = catalogues.__getattribute__(survey)
        survey_table = tables.__getattribute__(survey)

        if survey_catalogue is None:
            continue

        logger.info(f"Processing {survey=}")

        match_result = match_nearest_neighbour(
            table1=tables.askap,
            table2=survey_table,
            catalogue1=catalogues.askap,
            catalogue2=survey_catalogue,
        )

        xmatch_table, xmatch_file = make_xmatch_table(
            table1=tables.askap,
            table2=survey_table,
            catalogue1=catalogues.askap,
            catalogue2=survey_catalogue,
            match_result=match_result,
            output_path=output_path,
        )
        _tables[survey] = xmatch_file

    xmatch_tables = XMatchTables(
        **_tables,
    )

    stats_table = make_field_stats_table(
        field_summary=field_summary,
        rms_info=rms_info,
        output_path=output_path,
    )

    psf_table = make_psf_table(
        field_summary=field_summary,
        output_path=output_path,
    )

    return ValidationTables(
        xmatch_tables=xmatch_tables,
        stats_table_path=stats_table,
        psf_table_path=psf_table,
    )


def create_validation_plot(
    field_summary: FieldSummary,
    rms_image_path: Path,
    source_catalogue_path: Path,
    output_path: Path,
    reference_catalogue_directory: Path,
) -> Path:
    """Create a simple multi-panel validation figure intended to asses
    the correctness of an image and associated source catalogue.

    The image described by `rms_image_path` should be a FITS file. The
    WCS of this file is used for plotting and rreading the synthesised
    beam information using the standard CRVAL/BMAJ/BMIN keywords.

    The source catalogue is read using astropy.table.Table. This routine
    also expects that some level of units are embedded in the catalogue.
    For Aegean produced catalogues this is the case.

    The reference_catalogue_directory sets the directory to look into when
    searching for the reference ICRF, NVSS and SUMSS cataloues.

    Args:
        field_summary (FieldSummary): A description of the key properties of the field
        rms_image_path (Path): The RMS fits image the source catalogue was constructed against.
        source_catalogue_path (Path): The source catalogue.
        output_path (Path): The output path of the figure to create
        reference_catalogue_directory (Path): The directory that contains the reference ICRF, NVSS and SUMSS catalogues.

    Returns:
        Path: The output path of the figure
    """
    logger.info("Creating validation plot")

    rms_info = get_rms_image_info(rms_path=rms_image_path)

    dezotti_path = get_packaged_resource_path(
        package="flint.data.source_counts", filename="de_zotti_1p4.txt"
    )
    skads_path = get_packaged_resource_path(
        package="flint.data.source_counts", filename="SKADS_1p4GHz.fits"
    )
    logger.info(f"Loading {dezotti_path}")
    dezotti = Table.read(dezotti_path, format="ascii")
    logger.info(f"Loading {skads_path=}")
    skads = Table.read(skads_path)

    askap_survey_name = f"{field_summary.sbid}-{field_summary.field_name}"

    # Load in the catalogues
    catalogues, tables = load_catalogues(
        source_catalogue_path=source_catalogue_path,
        reference_catalogue_directory=reference_catalogue_directory,
        askap_survey_name=askap_survey_name,
        rms_info=rms_info,
    )

    height = 16.0
    width = height * np.sqrt(3.0)

    fig = plt.figure(figsize=(width, height))

    validator_layout = make_validator_axes_layout(fig=fig, rms_path=rms_image_path)

    plot_source_counts(
        catalogue=tables.askap,
        rms_info=rms_info,
        ax=validator_layout.ax_counts,
        freq=rms_info.header["CRVAL3"],
        dezotti=dezotti,
        skads=skads,
    )

    askap_sources = SkyCoord(
        tables.askap[catalogues.askap.ra_col],
        tables.askap[catalogues.askap.dec_col],
        unit="deg,deg",
    )
    plot_rms_map(
        fig=fig,
        ax=validator_layout.ax_rms,
        rms_path=rms_info.path,
        source_positions=askap_sources,
    )

    plot_psf(fig=fig, ax=validator_layout.ax_psf, rms_info=rms_info)

    ierf_match = match_nearest_neighbour(
        table1=tables.askap,
        table2=tables.icrf,
        catalogue1=catalogues.askap,
        catalogue2=catalogues.icrf,
    )
    plot_astrometry_comparison(
        fig=fig, ax=validator_layout.ax_astrometry, match_result=ierf_match
    )

    sumss_match = match_nearest_neighbour(
        table1=tables.askap,
        table2=tables.sumss,
        catalogue1=catalogues.askap,
        catalogue2=catalogues.sumss,
    )

    plot_astrometry_comparison(
        fig=fig, ax=validator_layout.ax_astrometry1, match_result=sumss_match
    )
    plot_flux_comparison(
        fig=fig, ax=validator_layout.ax_brightness1, match_result=sumss_match
    )

    nvss_match = match_nearest_neighbour(
        table1=tables.askap,
        table2=tables.nvss,
        catalogue1=catalogues.askap,
        catalogue2=catalogues.nvss,
    )

    plot_astrometry_comparison(
        fig=fig, ax=validator_layout.ax_astrometry2, match_result=nvss_match
    )
    plot_flux_comparison(
        fig=fig, ax=validator_layout.ax_brightness2, match_result=nvss_match
    )

    plot_field_info(
        fig=fig,
        ax=validator_layout.ax_legend,
        field_summary=field_summary,
        rms_info=rms_info,
        askap_table=tables.askap,
    )

    plot_flag_summary(
        fig=fig,
        ax=validator_layout.ax_flag_summary,
        field_summary=field_summary,
    )

    fig.suptitle(
        rf"$\it{{Flint}}$ summary for Field: {field_summary.field_name} - SBID: {field_summary.sbid} - Round: {field_summary.round}",
        fontsize=F_HUGE,
    )

    fig.tight_layout()
    output_file = output_path / f"validation_{askap_survey_name}.png"
    fig.savefig(str(output_file), dpi=300, bbox_inches="tight")

    return output_file


def get_parser() -> ArgumentParser:
    """Create the argument parser for the validation plot creation

    Returns:
        ArgumentParser: CLI entry point
    """
    parser = ArgumentParser(
        description="Create a validation figure to highlight the reliability of some process continuum field"
    )
    parser.add_argument(
        "processed_ms_paths",
        nargs="+",
        type=Path,
        help="Paths to the processed MeasurementSets.",
    )
    parser.add_argument(
        "rms_image_path", type=Path, help="Path to the RMS image of the field"
    )
    parser.add_argument(
        "source_catalogue_path",
        type=Path,
        help="Path to the source catalogue. At present on Aegean formatted component catalogues supported. ",
    )
    parser.add_argument(
        "output_path", type=Path, help="Directory to place the output files. "
    )
    parser.add_argument(
        "--reference-catalogue-directory",
        type=Path,
        default=Path("."),
        help="Directory container the reference ICFS, NVSS and SUMSS catalogues. These are known catalogues and expect particular file names. ",
    )
    parser.add_argument(
        "--bandpass-directory",
        type=Path,
        default=None,
        help="Path to the directory containing the bandpass measurement sets and solutions",
    )
    parser.add_argument(
        "--holography-path",
        type=Path,
        default=None,
        help="Path to the holography fits cube used to form the linmos image",
    )

    return parser


def cli() -> None:
    """CLI entry point for validation plot creation"""
    parser = get_parser()

    args = parser.parse_args()

    from astropy.io import fits

    from flint.source_finding.aegean import AegeanOutputs
    from flint.summary import create_field_summary

    rms_image_path = args.rms_image_path

    try:
        rms_header = fits.getheader(rms_image_path)
        rms_beam = (
            rms_header["BMAJ"],
            rms_header["BMIN"],
            rms_header["BPA"],
        )
    except KeyError:
        rms_beam = (1.0, 1.0, 1.0)
        logger.warning(
            f"Beam keywords not found in {rms_image_path=}. Setting to default {rms_beam}"
        )

    aegean_outputs = AegeanOutputs(
        bkg=rms_image_path,
        rms=rms_image_path,
        comp=args.source_catalogue_path,
        beam_shape=rms_beam,
        image=rms_image_path,
    )

    field_summary = create_field_summary(
        mss=args.processed_ms_paths,
        aegean_outputs=aegean_outputs,
        holography_path=args.holography_path,
        cal_sbid_path=args.bandpass_directory,
    )

    create_validation_plot(
        field_summary=field_summary,
        rms_image_path=args.rms_image_path,
        source_catalogue_path=args.source_catalogue_path,
        output_path=args.output_path,
        reference_catalogue_directory=args.reference_catalogue_directory,
    )

    validation_tables = create_validation_tables(
        field_summary=field_summary,
        rms_image_path=args.rms_image_path,
        source_catalogue_path=args.source_catalogue_path,
        output_path=args.output_path,
        reference_catalogue_directory=args.reference_catalogue_directory,
    )

    logger.info(f"\n{validation_tables}")


if __name__ == "__main__":
    cli()



"""Basic tests around prefect helper functions"""

from __future__ import annotations

from prefect import flow

from flint.prefect.helpers import enable_loguru_support


def test_enable_loguru_support():
    """Some packages may be using loguru (e.g. crystalball). Should
    we want those logs to be captured we need to modify the loguru
    logger. A helpful function has been added to this end. This
    is a simple, very basic test to make sure it can still run without
    error, though whether it still works is a completely different
    question!"""

    @flow
    def example_flow():
        enable_loguru_support()

    example_flow()



"""Tests around archives"""

from __future__ import annotations

import tarfile
from pathlib import Path

import pytest

from flint.archive import (
    copy_files_into,
    create_sbid_tar_archive,
    get_archive_options_from_yaml,
    get_parser,
    resolve_glob_expressions,
    tar_files_into,
)
from flint.options import DEFAULT_TAR_RE_PATTERNS, ArchiveOptions
from flint.utils import get_packaged_resource_path

FILES = [f"some_file_{a:02d}-MFS-image.fits" for a in range(36)] + [
    f"a_validation.{ext}" for ext in ("png", "jpeg", "pdf")
]


@pytest.fixture
def glob_files(tmpdir):
    """Create an example set of temporary files in a known directory"""
    for f in FILES:
        touch_file = f"{tmpdir / f!s}"
        with open(touch_file, "w") as out_file:
            out_file.write("example\n")

    return (tmpdir, FILES)


@pytest.fixture
def temp_files(glob_files):
    base_dir, files = glob_files

    archive_options = ArchiveOptions()

    resolved = resolve_glob_expressions(
        base_path=base_dir, file_re_patterns=archive_options.tar_file_re_patterns
    )

    return (base_dir, resolved)


@pytest.fixture
def package_strategy_path():
    example = get_packaged_resource_path(
        package="flint", filename="data/tests/test_config.yaml"
    )

    return example


def test_get_archive_options_from_yaml(package_strategy_path):
    """Get the options from a strategy yaml file for ArchiveOptions"""
    options = get_archive_options_from_yaml(
        strategy_yaml_path=Path(package_strategy_path)
    )

    assert isinstance(options, dict)
    assert options["tar_file_re_patterns"][-1] == "testing_for_jack.txt"
    assert options["copy_file_re_patterns"][-1] == "testing_for_sparrow.csv"
    assert len(options["tar_file_re_patterns"]) == 6


def test_copy_files_into(tmpdir, temp_files):
    """Basic sanity checks to the copying process"""
    base_dir, files = temp_files

    tmpdir = Path(tmpdir)
    copy_out = tmpdir / "copy_location"

    copy_files_into(copy_out_path=copy_out, files_to_copy=files)

    files_copied = list(copy_out.glob("*"))
    assert len(files_copied) == len(files)
    assert len(files_copied) > 0


def test_glob_expressions(glob_files):
    """Trying out the globbing from the archive options"""
    base_dir, files = glob_files

    archive_options = ArchiveOptions()
    assert len(archive_options.tar_file_re_patterns) > 0

    resolved = resolve_glob_expressions(
        base_path=base_dir, file_re_patterns=archive_options.tar_file_re_patterns
    )

    assert all([isinstance(p, Path) for p in resolved])
    assert len(resolved) == 37


def test_glob_expressions_uniq(glob_files):
    """Make sure that the uniqueness is correct"""
    base_dir, files = glob_files

    archive_options = ArchiveOptions(tar_file_re_patterns=(".*png", ".*png"))
    resolved = resolve_glob_expressions(
        base_path=base_dir, file_re_patterns=archive_options.tar_file_re_patterns
    )
    assert len(resolved) == 1


def test_glob_expressions_empty(glob_files):
    """Make sure that the uniqueness is correct"""
    base_dir, files = glob_files

    archive_options = ArchiveOptions(tar_file_re_patterns=(".*doesnotexist",))
    resolved = resolve_glob_expressions(
        base_path=base_dir, file_re_patterns=archive_options.tar_file_re_patterns
    )
    assert len(resolved) == 0


def test_archive_parser(glob_files):
    """Make sure pirates understand parsers"""
    base_dir, files = glob_files
    parser = get_parser()

    args = parser.parse_args("list".split())

    assert isinstance(args.base_path, Path)
    assert args.tar_file_re_patterns == DEFAULT_TAR_RE_PATTERNS

    example_path = Path("this/no/exist")
    args = parser.parse_args(f"list --base-path {example_path!s}".split())
    assert isinstance(args.base_path, Path)
    assert args.base_path == example_path

    args = parser.parse_args(
        r"list --copy-file-re-patterns '.*linmos.*' '.*MFS.*'".split()
    )
    assert len(args.copy_file_re_patterns) == 2

    example_path = Path(base_dir)
    args = parser.parse_args(
        f"list --base-path {example_path!s} --copy-file-re-patterns *pdf".split()
    )
    assert isinstance(args.base_path, Path)
    assert args.base_path == example_path
    assert args.copy_file_re_patterns == ["*pdf"]

    cmd = r"create --tar-file-re-patterns '.*linmos.*' '.*MFS.*' '.*beam[0-9]+\.round4-????-image\.fits' --base-path 39420 test_archive_tarball/39420.tar"
    args = parser.parse_args(cmd.split())
    assert len(args.tar_file_re_patterns) == 3


def test_tar_ball_files(temp_files):
    """Ensure that the tarballing works"""
    base_dir, files = temp_files

    tar_out_path = Path(base_dir) / "some_tarball.tar"
    _ = tar_files_into(tar_out_path=tar_out_path, files_to_tar=files)

    assert tarfile.is_tarfile(tar_out_path)

    with pytest.raises(FileExistsError):
        _ = tar_files_into(tar_out_path=tar_out_path, files_to_tar=files)


def test_create_sbid_archive(glob_files):
    """Attempts to ensure the entire find files and tarball creation works"""
    base_dir, files = glob_files

    archive_options = ArchiveOptions()
    tar_out_path = Path(base_dir) / "example_tarball_2.tar"

    _ = create_sbid_tar_archive(
        base_path=base_dir, tar_out_path=tar_out_path, archive_options=archive_options
    )

    assert tarfile.is_tarfile(tar_out_path)


def test_archive_new_tar_patterns():
    """Some sanity checks around this archive options tuple updating"""
    archive_options = ArchiveOptions()
    before_count = len(archive_options.tar_file_re_patterns)

    additional_file_patterns = (r".*beam[0-9]+\.round4-[0-9]{4}-image\.fits",)
    new_patterns = archive_options.tar_file_re_patterns + additional_file_patterns

    new_archive_options = archive_options.with_options(
        tar_file_re_patterns=new_patterns
    )
    assert len(new_archive_options.tar_file_re_patterns) == before_count + 1


def test_archiveoptions_with_options():
    """Ensure that the with_options interface for ArchiveOptions works"""

    archive_options = ArchiveOptions()
    default_copy = archive_options.copy_file_re_patterns
    update_options = ("Jack", "was", "here")
    new_options = archive_options.with_options(copy_file_re_patterns=update_options)

    assert new_options.copy_file_re_patterns != default_copy
    assert new_options.copy_file_re_patterns == update_options
    assert archive_options is not new_options

    new_dict = new_options._asdict()
    assert new_dict["copy_file_re_patterns"] == update_options



"""Very basic tests around the subtract model and cube work flow"""

from __future__ import annotations

from flint.prefect.flows.subtract_cube_pipeline import get_parser


def test_get_parser():
    """See the parser load successfully. Good test for imports and the like."""
    get_parser()



"""A prefect based pipeline that:
- will perform bandpass calibration with PKS B1934-638 data, or from a derived sky-model
- copy and apply to science field
- image and self-calibration the science fields
- run aegean source finding
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

from configargparse import ArgumentParser
from prefect import flow, tags, unmapped

from flint.calibrate.aocalibrate import find_existing_solutions
from flint.catalogue import verify_reference_catalogues
from flint.coadd.linmos import LinmosResult
from flint.configuration import (
    Strategy,
    get_options_from_strategy,
    load_and_copy_strategy,
)
from flint.logging import logger
from flint.masking import consider_beam_mask_round
from flint.ms import find_mss
from flint.naming import (
    CASDANameComponents,
    add_timestamp_to_path,
    extract_components_from_name,
    get_sbid_from_path,
)
from flint.options import (
    FieldOptions,
    add_options_to_parser,
    create_options_from_parser,
    dump_field_options_to_yaml,
)
from flint.prefect.clusters import get_dask_runner
from flint.prefect.common.imaging import (
    create_convol_linmos_images,
    create_convolve_linmos_cubes,
    task_copy_and_preprocess_casda_askap_ms,
    task_create_apply_solutions_cmd,
    task_create_image_mask_model,
    task_flag_ms_aoflagger,
    task_gaincal_applycal_ms,
    task_potato_peel,
    task_preprocess_askap_ms,
    task_rename_column_in_ms,
    task_run_bane_and_aegean,
    task_select_solution_for_ms,
    task_split_by_field,
    task_wsclean_imager,
    task_zip_ms,
    validation_items,
)
from flint.prefect.common.ms import task_add_model_source_list_to_ms
from flint.prefect.common.utils import (
    task_archive_sbid,
    task_create_beam_summary,
    task_create_field_summary,
    task_flatten,
    task_update_field_summary,
    task_update_with_options,
)
from flint.selfcal.utils import consider_skip_selfcal_on_round


def _check_field_options(field_options: FieldOptions) -> None:
    run_aegean = (
        False if field_options.aegean_container is None else field_options.run_aegean
    )
    if (
        field_options.imaging_strategy is not None
        and not field_options.imaging_strategy.exists()
    ):
        raise ValueError(
            f"Imaging strategy file {field_options.imaging_strategy} is set, but the path does not exist"
        )
    if field_options.use_beam_masks is True and run_aegean is False:
        raise ValueError(
            "run_aegean and aegean container both need to be set is beam masks is being used. "
        )
    if field_options.reference_catalogue_directory:
        if not verify_reference_catalogues(
            reference_directory=field_options.reference_catalogue_directory
        ):
            raise ValueError(
                f"{field_options.reference_catalogue_directory=} does not appear to be valid. Check for reference catalogues"
            )
    if field_options.rounds is not None:
        if field_options.rounds >= 1 and field_options.casa_container is None:
            raise ValueError(
                "CASA Container needs to be set if self-calibration is to be performed"
            )
    if field_options.coadd_cubes:
        if (
            field_options.yandasoft_container is None
            or not field_options.yandasoft_container
        ):
            raise ValueError(
                "Unable to create linmos cubes without a yandasoft container"
            )


def _check_create_output_split_science_path(
    science_path: Path, split_path: Path, check_exists: bool = True
) -> Path:
    """Create the output path that the science MSs will be placed.

    Args:
        science_path (Path): The directory that contains the MSs for science processing
        split_path (Path): Where the output MSs will be written to and processed
        check_exists (bool, optional): Should we check to make sure output directory does not exist. Defaults to True.

    Raises:
        ValueError: Raised when the output directory exists

    Returns:
        Path: The output directory
    """

    science_folder_name = science_path.name
    assert str(science_folder_name).isdigit(), (
        f"We require the parent directory to be the SBID (all digits), got {science_folder_name=}"
    )
    output_split_science_path = (
        Path(split_path / science_folder_name).absolute().resolve()
    )

    if check_exists and output_split_science_path.exists():
        logger.critical(
            f"{output_split_science_path=} already exists. It should not. Exiting. "
        )
        raise ValueError("Output science directory already exists. ")

    logger.info(f"Creating {output_split_science_path!s}")
    output_split_science_path.mkdir(parents=True)

    return output_split_science_path


@flow(name="Flint Continuum Pipeline")
def process_science_fields(
    science_path: Path,
    split_path: Path,
    field_options: FieldOptions,
    bandpass_path: Path | None = None,
) -> None:
    # Verify no nasty incompatible options
    _check_field_options(field_options=field_options)

    # Get some placeholder names
    run_aegean = (
        False if field_options.aegean_container is None else field_options.run_aegean
    )  # This is also in check_field_options
    run_validation = field_options.reference_catalogue_directory is not None

    science_mss = find_mss(
        mss_parent_path=science_path, expected_ms_count=field_options.expected_ms
    )

    output_split_science_path = _check_create_output_split_science_path(
        science_path=science_path, split_path=split_path, check_exists=True
    )

    dump_field_options_to_yaml(
        output_path=add_timestamp_to_path(
            input_path=output_split_science_path / "field_options.yaml"
        ),
        field_options=field_options,
    )

    archive_wait_for: list[Any] = []

    strategy: Strategy | None = load_and_copy_strategy(
        output_split_science_path=output_split_science_path,
        imaging_strategy=field_options.imaging_strategy,
    )

    logger.info(f"{field_options=}")

    logger.info(f"Found the following raw measurement sets: {science_mss}")

    # TODO: This feels a little too much like that feeling of being out
    # at sea for to long. Should refactor (or mask a EMU only).
    if isinstance(
        extract_components_from_name(name=science_mss[0].path), CASDANameComponents
    ):
        preprocess_science_mss = task_copy_and_preprocess_casda_askap_ms.map(
            casda_ms=science_mss, output_directory=output_split_science_path
        )
        preprocess_science_mss = task_flag_ms_aoflagger.map(  # type: ignore
            ms=preprocess_science_mss, container=field_options.flagger_container
        )
    else:
        # TODO: This will likely need to be expanded should any
        # other calibration strategies get added
        # Scan the existing bandpass directory for the existing solutions
        assert bandpass_path, f"{bandpass_path=}, it needs to be set"

        calibrate_cmds = find_existing_solutions(
            bandpass_directory=bandpass_path,
            use_preflagged=field_options.use_preflagger,
            use_smoothed=field_options.use_smoothed,
        )

        logger.info(f"Constructed the following {calibrate_cmds=}")

        split_science_mss = task_split_by_field.map(
            ms=science_mss,
            field=None,
            out_dir=unmapped(output_split_science_path),
            column=unmapped("DATA"),
        )

        # This will block until resolved
        flat_science_mss = task_flatten.submit(split_science_mss).result()

        solutions_paths = task_select_solution_for_ms.map(
            calibrate_cmds=unmapped(calibrate_cmds), ms=flat_science_mss
        )
        apply_solutions_cmds = task_create_apply_solutions_cmd.map(
            ms=flat_science_mss,
            solutions_file=solutions_paths,
            container=field_options.calibrate_container,
        )
        flagged_mss = task_flag_ms_aoflagger.map(
            ms=apply_solutions_cmds, container=field_options.flagger_container
        )
        column_rename_mss = task_rename_column_in_ms.map(
            ms=flagged_mss,
            original_column_name=unmapped("DATA"),
            new_column_name=unmapped("INSTRUMENT_DATA"),
        )
        preprocess_science_mss = task_preprocess_askap_ms.map(
            ms=column_rename_mss,
            data_column=unmapped("CORRECTED_DATA"),
            instrument_column=unmapped("DATA"),
            overwrite=True,
        )

    if field_options.no_imaging:
        logger.info(
            f"No imaging will be performed, as requested by {field_options.no_imaging=}"
        )
        return

    field_summary = task_create_field_summary.submit(
        mss=preprocess_science_mss,
        cal_sbid_path=bandpass_path,
        holography_path=field_options.holofile,
    )  # type: ignore
    logger.info(f"{field_summary=}")

    if field_options.wsclean_container is None:
        logger.info("No wsclean container provided. Rerutning. ")
        return

    if field_options.potato_container:
        # The call into potato peel task has two potential update option keywords.
        # So for the moment we will not use the task decorated version.
        potato_wsclean_init = get_options_from_strategy(
            strategy=strategy, mode="wsclean", round_info=0, operation="selfcal"
        )
        preprocess_science_mss = task_potato_peel.map(
            ms=preprocess_science_mss,
            potato_container=field_options.potato_container,
            update_wsclean_options=unmapped(potato_wsclean_init),
        )

    stokes_v_mss = preprocess_science_mss
    wsclean_results = task_wsclean_imager.map(
        in_ms=preprocess_science_mss,
        wsclean_container=field_options.wsclean_container,
        update_wsclean_options=unmapped(
            get_options_from_strategy(
                strategy=strategy,
                mode="wsclean",
                round_info=0,
                operation="selfcal",
            )
        ),
    )  # type: ignore

    wsclean_results = (
        task_add_model_source_list_to_ms.map(
            wsclean_command=wsclean_results,
            calibrate_container=field_options.calibrate_container,
        )
        if field_options.update_model_data_with_source_list
        else wsclean_results
    )

    # TODO: This should be waited!
    beam_summaries = task_create_beam_summary.map(
        ms=preprocess_science_mss, image_set=wsclean_results
    )
    archive_wait_for.extend(beam_summaries)
    archive_wait_for.extend(wsclean_results)

    beam_aegean_outputs = None
    if run_aegean:
        update_bane_options = get_options_from_strategy(
            strategy=strategy,
            mode="bane",
            round_info=0,
            operation="selfcal",
        )
        update_aegean_options = get_options_from_strategy(
            strategy=strategy,
            mode="aegean",
            round_info=0,
            operation="selfcal",
        )
        beam_aegean_outputs = task_run_bane_and_aegean.map(
            image=wsclean_results,
            aegean_container=unmapped(field_options.aegean_container),
            update_bane_options=unmapped(update_bane_options),
            update_aegean_options=unmapped(update_aegean_options),
        )
        beam_summaries = task_update_with_options.map(
            input_object=beam_summaries, components=beam_aegean_outputs
        )
        field_summary = task_update_with_options.submit(
            input_object=field_summary, beam_summaries=beam_summaries
        )

    if field_options.yandasoft_container:
        parsets = create_convol_linmos_images(
            wsclean_results=wsclean_results,
            field_options=field_options,
            field_summary=field_summary,
            additional_linmos_suffix_str="noselfcal",  # indicate in output linmos name no selfcal
        )
        archive_wait_for.extend(parsets)
        parset = parsets[-1]

        if run_aegean:
            aegean_field_output = task_run_bane_and_aegean.submit(
                image=parset, aegean_container=unmapped(field_options.aegean_container)
            )  # type: ignore
            field_summary = task_update_field_summary.submit(
                field_summary=field_summary,
                aegean_outputs=aegean_field_output,
                linmos_command=parset,
            )  # type: ignore
            archive_wait_for.append(field_summary)

            if run_validation and field_options.reference_catalogue_directory:
                validation_items(
                    field_summary=field_summary,
                    aegean_outputs=aegean_field_output,
                    reference_catalogue_directory=field_options.reference_catalogue_directory,
                )

    # Set up the default value should the user activated mask option is not set
    fits_beam_masks = None

    for current_round in range(1, field_options.rounds + 1):
        with tags(f"selfcal-{current_round}"):
            final_round = current_round == field_options.rounds

            skip_gaincal_current_round = consider_skip_selfcal_on_round(
                current_round=current_round,
                skip_selfcal_on_rounds=field_options.skip_selfcal_on_rounds,
            )

            update_gain_options = get_options_from_strategy(
                strategy=strategy,
                mode="gaincal",
                round_info=current_round,
                operation="selfcal",
            )
            cal_mss = task_gaincal_applycal_ms.map(
                ms=wsclean_results,
                selfcal_round=current_round,
                archive_input_ms=field_options.zip_ms,
                skip_selfcal=skip_gaincal_current_round,
                rename_ms=field_options.rename_ms,
                archive_cal_table=True,
                casa_container=field_options.casa_container,
                update_gain_cal_options=unmapped(update_gain_options),
                wait_for=[
                    field_summary
                ],  # To make sure field summary is created with unzipped MSs
            )  # type: ignore
            stokes_v_mss = cal_mss

            fits_beam_masks = None
            if consider_beam_mask_round(
                current_round=current_round,
                mask_rounds=(
                    field_options.use_beam_masks_rounds
                    if field_options.use_beam_masks_rounds
                    else field_options.use_beam_masks_from
                ),
                allow_beam_masks=field_options.use_beam_masks,
            ):
                # TODO: the aegean outputs are only needed should
                # the signal image be needed or per beam sources are needed
                update_bane_options = get_options_from_strategy(
                    strategy=strategy,
                    mode="bane",
                    round_info=current_round,
                    operation="selfcal",
                )
                update_aegean_options = get_options_from_strategy(
                    strategy=strategy,
                    mode="aegean",
                    round_info=current_round,
                    operation="selfcal",
                )

                beam_aegean_outputs = (
                    task_run_bane_and_aegean.map(
                        image=wsclean_results,
                        aegean_container=unmapped(field_options.aegean_container),
                        update_bane_options=unmapped(update_bane_options),
                        update_aegean_options=unmapped(update_aegean_options),
                    )
                    if (current_round >= 2 or not beam_aegean_outputs)
                    else beam_aegean_outputs
                )
                # Early versions of the masking procedure required aegean outputs
                # to construct the sginal images. Since aegean is run outside of
                # this self-cal loop once already, we can skip their creation on
                # the first loop

                update_masking_options = get_options_from_strategy(
                    strategy=strategy,
                    mode="masking",
                    round_info=current_round,
                    operation="selfcal",
                )
                fits_beam_masks = task_create_image_mask_model.map(
                    image=wsclean_results,
                    image_products=beam_aegean_outputs,
                    update_masking_options=unmapped(update_masking_options),
                )  # type: ignore

            update_wsclean_options = get_options_from_strategy(
                strategy=strategy,
                mode="wsclean",
                operation="selfcal",
                round_info=current_round,
            )
            wsclean_results = task_wsclean_imager.map(
                in_ms=cal_mss,
                wsclean_container=field_options.wsclean_container,
                fits_mask=fits_beam_masks,
                update_wsclean_options=unmapped(update_wsclean_options),
            )
            wsclean_results = (
                task_add_model_source_list_to_ms.map(
                    wsclean_command=wsclean_results,
                    calibrate_container=field_options.calibrate_container,
                )
                if field_options.update_model_data_with_source_list
                else wsclean_results
            )
            archive_wait_for.extend(wsclean_results)

            # Do source finding on the last round of self-cal'ed images
            if round == field_options.rounds and run_aegean:
                task_run_bane_and_aegean.map(
                    image=wsclean_results,
                    aegean_container=unmapped(field_options.aegean_container),
                )

            parsets_self: None | list[LinmosResult] = None  # Without could be unbound
            if field_options.yandasoft_container:
                parsets_self = create_convol_linmos_images(
                    wsclean_results=wsclean_results,
                    field_options=field_options,
                    field_summary=field_summary,
                )
                archive_wait_for.extend(parsets_self)

            if final_round and run_aegean and parsets_self:
                aegean_outputs = task_run_bane_and_aegean.submit(
                    image=parsets_self[-1],
                    aegean_container=unmapped(field_options.aegean_container),
                )  # type: ignore
                field_summary = task_update_field_summary.submit(
                    field_summary=field_summary,
                    aegean_outputs=aegean_outputs,
                    round=current_round,
                )  # type: ignore
                if run_validation:
                    assert field_options.reference_catalogue_directory, (
                        f"Reference catalogue directory should be set when {run_validation=}"
                    )
                    val_results = validation_items(
                        field_summary=field_summary,
                        aegean_outputs=aegean_outputs,
                        reference_catalogue_directory=field_options.reference_catalogue_directory,
                    )
                    archive_wait_for.append(val_results)

    if field_options.coadd_cubes:
        with tags("cubes"):
            cube_parset = create_convolve_linmos_cubes(
                wsclean_results=wsclean_results,  # type: ignore
                field_options=field_options,
                current_round=(field_options.rounds if field_options.rounds else None),
                additional_linmos_suffix_str="cube",
            )
            archive_wait_for.append(cube_parset)

    if field_options.stokes_v_imaging:
        with tags("stokes-v"):
            stokes_v_wsclean_options = get_options_from_strategy(
                strategy=strategy, mode="wsclean", operation="stokesv"
            )
            wsclean_results = task_wsclean_imager.map(
                in_ms=stokes_v_mss,
                wsclean_container=field_options.wsclean_container,
                update_wsclean_options=unmapped(stokes_v_wsclean_options),
                fits_mask=fits_beam_masks,
                wait_for=wsclean_results,  # Ensure that measurement sets are doubled up during imaging
            )
            if field_options.yandasoft_container:
                parsets = create_convol_linmos_images(
                    wsclean_results=wsclean_results,
                    field_options=field_options.with_options(linmos_residuals=False),
                    field_summary=field_summary,
                )
                archive_wait_for.extend(parsets)

    # zip up the final measurement set, which is not included in the above loop
    if field_options.zip_ms:
        archive_wait_for = task_zip_ms.map(
            in_item=wsclean_results, wait_for=archive_wait_for
        )

    if field_options.sbid_archive_path or field_options.sbid_copy_path:
        update_archive_options = get_options_from_strategy(
            strategy=strategy, mode="archive", round_info=0, operation="selfcal"
        )
        task_archive_sbid.submit(
            science_folder_path=output_split_science_path,
            archive_path=field_options.sbid_archive_path,
            copy_path=field_options.sbid_copy_path,
            max_round=field_options.rounds if field_options.rounds else None,
            update_archive_options=update_archive_options,
            wait_for=archive_wait_for,
        )


def setup_run_process_science_field(
    cluster_config: str | Path,
    science_path: Path,
    split_path: Path,
    field_options: FieldOptions,
    bandpass_path: Path | None = None,
) -> None:
    science_sbid = get_sbid_from_path(path=science_path)

    if field_options.sbid_copy_path:
        updated_sbid_copy_path = field_options.sbid_copy_path / f"{science_sbid}"
        logger.info(f"Updating archive copy path to {updated_sbid_copy_path=}")
        field_options = field_options.with_options(
            sbid_copy_path=updated_sbid_copy_path
        )

    dask_task_runner = get_dask_runner(cluster=cluster_config)

    process_science_fields.with_options(
        name=f"Flint Continuum Pipeline - {science_sbid}", task_runner=dask_task_runner
    )(
        science_path=science_path,
        bandpass_path=bandpass_path,
        split_path=split_path,
        field_options=field_options,
    )

    # TODO: Put the archive stuff here?


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)

    parser.add_argument(
        "--cli-config", is_config_file=True, help="Path to configuration file"
    )

    parser.add_argument(
        "science_path",
        type=Path,
        help="Path to directories containing the beam-wise science measurementsets that will have solutions copied over and applied.",
    )
    parser.add_argument(
        "--split-path",
        type=Path,
        default=Path("."),
        help="Location to write field-split MSs to. Will attempt to use the parent name of a directory when writing out a new MS. ",
    )
    parser.add_argument(
        "--calibrated-bandpass-path",
        type=Path,
        default=None,
        help="Path to directory containing the uncalibrated beam-wise measurement sets that contain the bandpass calibration source. If None then the '--sky-model-directory' should be provided. ",
    )
    parser.add_argument(
        "--cluster-config",
        type=str,
        default="petrichor",
        help="Path to a cluster configuration file, or a known cluster name. ",
    )

    parser = add_options_to_parser(parser=parser, options_class=FieldOptions)

    return parser


def cli() -> None:
    import logging

    # logger = logging.getLogger("flint")
    logger.setLevel(logging.INFO)

    parser = get_parser()

    args = parser.parse_args()

    field_options: FieldOptions = create_options_from_parser(
        parser_namespace=args,
        options_class=FieldOptions,
    )

    setup_run_process_science_field(
        cluster_config=args.cluster_config,
        science_path=args.science_path,
        bandpass_path=args.calibrated_bandpass_path,
        split_path=args.split_path,
        field_options=field_options,
    )


if __name__ == "__main__":
    cli()



from __future__ import annotations

import pytest

from flint.logging import logger


# Stolen from: https://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
def which(program: str) -> str | None:
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ.get("PATH", "").split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


if which("singularity") is None:
    pytest.skip("Singularity is not installed", allow_module_level=True)


def test_singularity():
    which_singularity = which("singularity")
    logger.info(f"Singularity is installed at: {which_singularity}")
    assert which_singularity is not None



"""Testing components in the leakage creation steps"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from astropy.table import Table
from astropy.wcs import WCS

from flint.leakage import (
    FITSImage,
    LeakageFilters,
    PixelCoords,
    _get_output_catalogue_path,
    _load_component_table,
    _load_fits_image,
    filter_components,
    get_xy_pixel_coords,
)
from flint.utils import get_packaged_resource_path


def test_get_catalogue_output_path():
    """Make sure the name of the catalogue comes out as expected"""
    catalogue = Path("SB1234.JACK_1234+56.i.round2_comp.fits")
    pol = "v"

    output_path = _get_output_catalogue_path(input_path=catalogue, pol=pol)
    assert output_path == Path("SB1234.JACK_1234+56.i.round2_comp.v_leakage.fits")

    spec_path = Path("JackSparrow_leakage.fits")
    output_path = _get_output_catalogue_path(
        input_path=catalogue, pol=pol, output_path=spec_path
    )
    assert output_path == spec_path

    with pytest.raises(AssertionError):
        _get_output_catalogue_path(input_path="notapathjack", pol=pol)


def _get_aegean_catalogue():
    table_path = get_packaged_resource_path(
        package="flint.data.tests",
        filename="SB38959.RACS_1357-18.noselfcal.linmos_comp.fits",
    )
    table = Table.read(table_path)
    return table, table_path


def _get_fits_image():
    fits_path = get_packaged_resource_path(
        package="flint.data.tests",
        filename="SB56659.RACS_0940-04.beam17.round3-0000-image.sub.fits",
    )

    return fits_path


def test_get_xy_coords():
    """Test on the conversion of component positions to pixel (x,y)"""
    fits_path = _get_fits_image()
    table, _ = _get_aegean_catalogue()

    fits_image = _load_fits_image(fits_path=fits_path)
    pixel_coords = get_xy_pixel_coords(table=table, wcs=fits_image.wcs)

    assert isinstance(pixel_coords, PixelCoords)
    assert pixel_coords.x.dtype == int
    assert pixel_coords.y.dtype == int
    assert len(pixel_coords.x) == len(pixel_coords.y)
    assert len(pixel_coords.x) == len(table)


def test_filter_component_catalogue():
    """Attempt to filter components from a catalogue"""
    table, _ = _get_aegean_catalogue()

    leakage_filters = LeakageFilters(
        isolation_radius_deg=0.0155,
        upper_int_peak_ratio=1.2,
        lower_int_peak_ratio=0.8,
        source_snr=40,
    )

    subset_table = filter_components(
        table=table,
        peak_col="peak_flux",
        int_col="int_flux",
        int_err_col="local_rms",
        leakage_filters=leakage_filters,
    )
    assert (
        len(subset_table) == 523
    )  # Did this in an interactive shell to get the expected.

    with pytest.raises(AssertionError):
        filter_components(
            table=table,
            peak_col="JackSparrow",
            int_col="int_flux",
            int_err_col="local_rms",
            leakage_filters=leakage_filters,
        )


def test_load_component_catalogue():
    """Test loading a component catalogue"""
    table, table_path = _get_aegean_catalogue()
    assert isinstance(table_path, Path)

    return_table = _load_component_table(catalogue=table)
    assert return_table is table

    loaded_table = _load_component_table(catalogue=table_path)
    assert loaded_table is not Table
    assert loaded_table.colnames == return_table.colnames


def test_load_fits_image():
    """Testinging the load function that attempts to read a FITS image and extract its bits"""
    fits_path = _get_fits_image()

    assert isinstance(fits_path, Path)
    assert fits_path.exists()

    fits_image = _load_fits_image(fits_path=fits_path)

    assert isinstance(fits_image, FITSImage)
    assert fits_image.path == fits_path
    assert isinstance(fits_image.data, np.ndarray)
    assert isinstance(fits_image.wcs, WCS)
    assert isinstance(fits_image.header, dict)

    with pytest.raises(AssertionError):
        _load_fits_image(fits_path=fits_path.with_suffix(".jack.sparrow"))



"""Utilities related to using casa tasks"""

from __future__ import annotations

from pathlib import Path

from flint.logging import logger
from flint.sclient import singularity_wrapper
from flint.utils import remove_files_folders


def args_to_casa_task_string(task: str, **kwargs) -> str:
    """Given a set of arguments, convert them to a string that can
    be used to run the corresponding CASA task that can be passed
    via ``casa -c`` for execution

    Args:
        task (str): The name of the task that will be executed

    Returns:
        str: The formatted string that will be given to CASA for execution
    """
    command = []
    for k, v in kwargs.items():
        if isinstance(v, (list, tuple)):
            v = ",".join(rf"'{_v!s}'" for _v in v)
            arg = rf"{k}=({v})"
        elif isinstance(v, (str, Path)):
            arg = rf"{k}='{v!s}'"
        else:
            arg = rf"{k}={v}"
        command.append(arg)

    task_command = rf"casa -c {task}(" + ",".join(command) + r")"

    return task_command


# TODO There should be a general casa_command type function that accepts the task as a keyword
# so that each casa task does not need an extra function


@singularity_wrapper
def mstransform(**kwargs) -> str:
    """Construct and run CASA's ``mstransform`` task.

    Args:
        casa_container (Path): Container with the CASA tooling
        ms (str): Path to the measurement set to transform
        output_ms (str): Path of the output measurement set produced by the transform

    Returns:
        str: The ``mstransform`` string
    """
    mstransform_str = args_to_casa_task_string(task="mstransform", **kwargs)
    logger.info(f"{mstransform_str=}")

    return mstransform_str


@singularity_wrapper
def cvel(**kwargs) -> str:
    """Generate the CASA cvel command

    Returns:
        str: The command to execute
    """
    cvel_str = args_to_casa_task_string(task="cvel", **kwargs)
    logger.info(f"{cvel_str=}")

    return cvel_str


@singularity_wrapper
def applycal(**kwargs) -> str:
    """Generate the CASA applycal command

    Returns:
        str: The command to execute
    """
    applycal_str = args_to_casa_task_string(task="applycal", **kwargs)
    logger.info(f"{applycal_str=}")

    return applycal_str


@singularity_wrapper
def gaincal(**kwargs) -> str:
    """Generate the CASA gaincal command

    Returns:
        str: The command to execute
    """
    gaincal_str = args_to_casa_task_string(task="gaincal", **kwargs)
    logger.info(f"{gaincal_str=}")

    return gaincal_str


def copy_with_mstranform(casa_container: Path, ms_path: Path) -> Path:
    """Use the casa task mstransform to create `nspw` spectral windows
    in the input measurement set. This is necessary when attempting to
    use gaincal to solve for some frequency-dependent solution.

    Args:
        casa_container (Path): Path to the singularity container with CASA tooling
        ms_path (Path): The measurement set that should be reformed to have `nspw` spectral windows

    Returns:
        Path: The path to the measurement set that was updated
    """

    transform_ms = ms_path.with_suffix(".ms_transform")

    mstransform(
        container=casa_container,
        bind_dirs=(ms_path.parent, transform_ms.parent),
        vis=str(ms_path),
        outputvis=str(transform_ms),
        createmms=False,
        datacolumn="all",
    )

    logger.info(f"Successfully created the transformed measurement set {transform_ms}.")
    remove_files_folders(ms_path)

    logger.info(f"Renaming {transform_ms} to {ms_path}.")
    transform_ms.rename(ms_path)

    return ms_path



"""Tests around the sky-model code"""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest

from flint.options import create_options_from_parser
from flint.sky_model import (
    SkyModel,
    SkyModelOptions,
    SkyModelOutputPaths,
    create_sky_model,
    get_parser,
    get_sky_model_output_paths,
)
from flint.utils import get_packaged_resource_path


@pytest.fixture
def ms_example_and_nvss(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="SB39400.RACS_0635-31.beam0.small.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "39400"

    shutil.unpack_archive(ms_zip, outpath)

    nvss_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="NVSS.fits.zip",
        )
    )

    shutil.unpack_archive(nvss_zip, outpath)

    ms_path = Path(outpath) / "SB39400.RACS_0635-31.beam0.small.ms"
    nvss_path = Path(outpath) / "NVSS.fits"

    return ms_path, nvss_path


def test_extracting_sky_model(ms_example_and_nvss):
    """Run the whole sky model creating to make sure nothing has
    broken"""
    ms_path, nvss_path = ms_example_and_nvss
    assert ms_path.exists()
    assert nvss_path.exists()

    sky_model_options = SkyModelOptions(
        reference_catalogue_directory=nvss_path.parent,
        reference_name="NVSS",
    )
    sky_model = create_sky_model(ms_path=ms_path, sky_model_options=sky_model_options)

    assert isinstance(sky_model, SkyModel)
    assert sky_model.no_sources == 18
    assert sky_model.calibrate_model is None
    assert sky_model.hyperdrive_model is None
    assert sky_model.ds9_region is None

    sky_model_options = SkyModelOptions(
        reference_catalogue_directory=nvss_path.parent,
        reference_name="NVSS",
        write_calibrate_model=True,
        write_hyperdrive_model=True,
        write_ds9_region=True,
    )
    sky_model = create_sky_model(ms_path=ms_path, sky_model_options=sky_model_options)

    assert isinstance(sky_model, SkyModel)
    assert sky_model.no_sources == 18
    assert (
        isinstance(sky_model.calibrate_model, Path)
        and sky_model.calibrate_model.exists()
    )
    assert (
        isinstance(sky_model.hyperdrive_model, Path)
        and sky_model.hyperdrive_model.exists()
    )
    assert isinstance(sky_model.ds9_region, Path) and sky_model.ds9_region.exists()
    assert np.isclose(sky_model.flux_jy, 1.3768, rtol=0.01)


def test_extracting_sky_model_with_none(ms_example_and_nvss):
    """Run the whole sky model creating to make sure nothing has
    broken. Make sure though that no sources are reported and a
    None is returned instead"""
    ms_path, nvss_path = ms_example_and_nvss
    assert ms_path.exists()
    assert nvss_path.exists()

    sky_model_options = SkyModelOptions(
        reference_catalogue_directory=nvss_path.parent,
        reference_name="NVSS",
        flux_cutoff=100000,
    )
    sky_model = create_sky_model(ms_path=ms_path, sky_model_options=sky_model_options)

    assert sky_model is None


def test_get_working_parser():
    """Make sure that the interaction with the SkyModelOptions
    and the argument parser works"""
    parser = get_parser()

    args = parser.parse_args("example.ms".split())
    ms = Path(args.ms)
    assert ms == Path("example.ms")

    sky_model_options = create_options_from_parser(
        parser_namespace=args, options_class=SkyModelOptions
    )
    assert isinstance(sky_model_options, SkyModelOptions)
    assert isinstance(sky_model_options.reference_catalogue_directory, Path)


def test_get_sky_model_output_names():
    """Ensure the names are what we expect them to be"""

    ms_path = Path("JackSparrowData.ms")

    sky_model_output_paths = get_sky_model_output_paths(ms_path=ms_path)
    assert isinstance(sky_model_output_paths, SkyModelOutputPaths)
    assert sky_model_output_paths.hyperdrive_path == Path(
        "JackSparrowData.hyperdrive.yaml"
    )
    assert sky_model_output_paths.calibrate_path == Path(
        "JackSparrowData.calibrate.txt"
    )

    with pytest.raises(ValueError):
        get_sky_model_output_paths(ms_path=Path("JackBeNoMS.txt"))



"""Utilities related to using casa tasks to perform self-calibration.

This tooling is mostly centred on using gaincal from casatasks.
"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from shutil import copytree
from typing import Any

from casacore.tables import table

from flint.casa import applycal, cvel, gaincal, mstransform
from flint.exceptions import GainCalError, MSError
from flint.flagging import nan_zero_extreme_flag_ms
from flint.logging import logger
from flint.ms import rename_ms_and_columns_for_selfcal
from flint.naming import get_selfcal_ms_name
from flint.options import MS, BaseOptions
from flint.selfcal.utils import (
    create_and_check_caltable_path,
    get_channel_ranges_given_nspws_for_ms,
)
from flint.utils import remove_files_folders, rsync_copy_directory, zip_folder


class GainCalOptions(BaseOptions):
    """Options provided to the casatasks gaincal function. Most options correspond to those in gaincal."""

    solint: str = "60s"
    """Solution length interval"""
    calmode: str = "p"
    """Calibration mode. """
    round: int = 0
    """Self-calibration round. Not a gaincal option. """
    minsnr: float = 0.0
    """Minimum signal-to-noise of the solutions. Below this the solutions and data are flagged. """
    uvrange: str = ">200m"
    """Data selected to go through calibration procedure"""
    selectdata: bool = True
    """Whether data selection actions will be applied in gaincal. """
    gaintype: str = "G"
    """The gain type that would be solved for. """
    nspw: int = 1
    """The number of spectral windows to use during the self-calibration routine. This
    will be used to craft an appropriate ``select_spw=`` interval range. If larger
    than one, ``gaincal`` will be carried out against each interval and results will
    be appended to a common solutions file. """


def copy_and_clean_ms_casagain(
    ms: MS, round: int = 1, verify: bool = True, rename_ms: bool = False
) -> MS:
    """Create a copy of a measurement set in preparation for selfcalibration
    using casa's gaincal and applycal. Applycal only works when calibrating
    DATA and creating a CORRECTED_DATA column. Columns are removed in the
    copied MS to allow this.

    If the MS is large the `move_ms` option will simply renamed the input MS, adjusting
    the columns appropriately. Note that this potentially involves deleting the `DATA`
    column if present and renaming `CORRECTED_DATA`.

    Args:
        ms (MS): Measurement set that would go through self-calibration.
        round (int, optional): The self-calibration round. Defaults to 1.
        verify (bool, optional): Verify that copying the measurementt set (done in preparation for gaincal) was successful. Uses a call to rsync. Defaults to True.
        move_ms (bool, optional): Rather than copying the MS, simple renamed the MS and modify columns appropriately. Defaults to False.

    Returns:
        MS: Copy of input measurement set with columns removed as required.
    """
    out_ms_path = get_selfcal_ms_name(in_ms_path=ms.path, round=round)

    mode_text = "Renaming" if rename_ms else "Copying"

    logger.info(f"Output MS name will be {out_ms_path!s}.")
    logger.info(f"{mode_text} {ms.path} to {out_ms_path}.")

    if out_ms_path.exists():
        logger.warning(f"{out_ms_path} already exists. Removing it. ")
        logger.warning(f"{ms.path=} {ms.path.exists()=}")
        logger.warning(f"{out_ms_path=} {out_ms_path.exists()=}")
        remove_files_folders(out_ms_path)

    if rename_ms:
        if not ms.column:
            raise MSError(f"No column has been assigned: {ms}")

        ms = rename_ms_and_columns_for_selfcal(
            ms=ms, target=out_ms_path, corrected_data=ms.column, data="DATA"
        )
    else:
        copytree(ms.path, out_ms_path)
        # Because we can trust nothing, verify the
        # copy with rsync. On some lustre file systems (mostly seen on stonix)
        # components of the measurement sett are not always successfully
        # copied with a simple copytree.
        if verify:
            rsync_copy_directory(ms.path, out_ms_path)

        logger.info("Copying finished. ")

        # The casa gaincal and applycal tasks __really__ expect the input and output
        # column names to be DATA and CORRECTED_DATA. So, here we will go through the
        # motions of deleting and rnaming columns. Note that the MODEL_DATA column needs
        # to exist. The INSTRUMENT_DATA column will also be removed.
        logger.info("About to open the table. ")
        with table(str(out_ms_path), readonly=False, ack=False) as tab:
            logger.info("About to get the colnames")
            colnames = tab.colnames()
            logger.info(f"Column names are: {colnames}")
            if ms.column == "DATA" and "CORRECTED_DATA" not in colnames:
                logger.info(
                    "Data is the nominated column, and CORRECTED_DATA does not exist. Returning. "
                )
            else:
                to_delete = [
                    "DATA",
                ]
                for col in to_delete:
                    if col in colnames:
                        logger.info(f"Removing {col=} from {out_ms_path!s}.")
                        try:
                            tab.removecols(col)
                            tab.flush(recursive=True)
                        except Exception as e:
                            logger.critical(
                                f"Failed to remove {col=}! \nCaptured error: {e}"
                            )
                    else:
                        logger.warning(f"Column {col} not found in {out_ms_path!s}.")

                logger.info("Renaming CORRECTED_DATA to DATA. ")
                tab.renamecol("CORRECTED_DATA", "DATA")

            tab.flush()

    # Note that the out_ms_path needs to be set, even if the data  column is initially DATA.
    # Since casa expects DATA, we will force the column to be DATA with the expectation that
    # previous pirates in the lines above have dealt with the renaming.
    ms = ms.with_options(path=out_ms_path, column="DATA")

    ms = nan_zero_extreme_flag_ms(ms=ms)

    return ms


def create_spws_in_ms(casa_container: Path, ms_path: Path, nspw: int) -> Path:
    """Use the casa task mstransform to create `nspw` spectral windows
    in the input measurement set. This is necessary when attempting to
    use gaincal to solve for some frequency-dependent solution.

    Args:
        casa_container (Path): Path to the singularity container with CASA tooling
        ms_path (Path): The measurement set that should be reformed to have `nspw` spectral windows
        nspw (int): The number of spectral windows to create

    Returns:
        Path: The path to the measurement set that was updated
    """

    logger.info(f"Transforming {ms_path!s} to have {nspw} SPWs")
    transform_ms = ms_path.with_suffix(".ms_transform")

    mstransform(
        container=casa_container,
        bind_dirs=(ms_path.parent, transform_ms.parent),
        vis=str(ms_path),
        outputvis=str(transform_ms),
        regridms=True,
        nspw=nspw,
        mode="channel",
        nchan=-1,
        start=0,
        width=1,
        chanbin=1,
        createmms=False,
        datacolumn="all",
        combinespws=False,
    )

    logger.info(
        f"Successfully created the transformed measurement set {transform_ms} with {nspw} SPWs."
    )
    remove_files_folders(ms_path)

    logger.info(f"Renaming {transform_ms} to {ms_path}.")
    transform_ms.rename(ms_path)

    # Look above - we have renamed the cvel measurement set Captain
    return ms_path


def merge_spws_in_ms(casa_container: Path, ms_path: Path) -> Path:
    """Attempt to merge together all SPWs in the input measurement
    set using the `cvel` casa task. This can be a little fragile.

    The `cvel` task creates a new measurement set, so there would
    temporially be a secondary measurement set.

    Args:
        ms_path (Path): The measurement set that should have its SPWs merged together.

    Returns:
        Path: The measurement set with merged SPWs. It will have the same path as `ms_path`, but is a new measurement set.
    """
    logger.info("Will attempt to merge all spws using cvel.")

    cvel_ms_path = ms_path.with_suffix(".cvel")
    cvel(
        container=casa_container,
        bind_dirs=(ms_path.parent,),
        vis=str(ms_path),
        outputvis=str(cvel_ms_path),
        mode="channel_b",
    )

    logger.info(f"Successfully merged spws in {cvel_ms_path}")

    remove_files_folders(ms_path)

    logger.info(f"Renaming {cvel_ms_path} to {ms_path}")
    cvel_ms_path.rename(ms_path)

    # Look above - we have renamed the cvel measurement set Captain
    return ms_path


def gaincal_applycal_ms(
    ms: MS,
    casa_container: Path,
    round: int = 1,
    gain_cal_options: GainCalOptions | None = None,
    update_gain_cal_options: dict[str, Any] | None = None,
    archive_input_ms: bool = False,
    raise_error_on_fail: bool = True,
    skip_selfcal: bool = False,
    rename_ms: bool = False,
    archive_cal_table: bool = False,
) -> MS:
    """Perform self-calibration using casa's gaincal and applycal tasks against
    an input measurement set.

    Args:
        ms (MS): Measurement set that will be self-calibrated.
        round (int, optional): Round of self-calibration, which is used for unique names. Defaults to 1.
        casa_container (Path): A path to a singularity container with CASA tooling.
        gain_cal_options (Optional[GainCalOptions], optional): Options provided to gaincal. Defaults to None.
        update_gain_cal_options (Optional[Dict[str, Any]], optional): Update the gain_cal_options with these. Defaults to None.
        archive_input_ms (bool, optional): If True, the input measurement set will be compressed into a single file. Defaults to False.
        raise_error_on_fail (bool, optional): If gaincal does not converge raise en error. If False and gain cal fails return the input ms. Defaults to True.
        skip_selfcal (bool, optional): Should this self-cal be skipped. If `True`, the a new MS is created but not calibrated the appropriate new name and returned.
        rename_ms (bool, optional): It `True` simply rename a MS and adjust columns appropriately (potentially deleting them) instead of copying the complete MS. If `True` `archive_input_ms` is ignored. Defaults to False.
        archive_cal_table (bool, optional): Archive the output calibration table in a tarball. Defaults to False.

    Raises:
        GainCallError: Raised when raise_error_on_fail is True and gaincal does not converge.

    Returns:
        MS: The self-calibrated measurement set.
    """
    logger.info(f"Measurement set to be self-calibrated: ms={ms}")

    assert casa_container.exists(), f"{casa_container=} does not exist. "

    if gain_cal_options is None:
        gain_cal_options = GainCalOptions()
    if update_gain_cal_options:
        logger.info(f"Updating gaincal options with: {update_gain_cal_options}")
        gain_cal_options = gain_cal_options.with_options(**update_gain_cal_options)

    # TODO: If the skip_selfcal is True we should just symlink, maybe?
    # Pirates like easy things though.
    cal_ms = copy_and_clean_ms_casagain(ms=ms, round=round, rename_ms=rename_ms)

    # Archive straight after copying in case we skip the gaincal and return
    if archive_input_ms and not rename_ms:
        zip_folder(in_path=ms.path)

    # No need to do work me, hardy
    if skip_selfcal:
        logger.info(f"{skip_selfcal=}, not calibrating the MS. ")
        return cal_ms

    # First, we collect the solutions for each of the requested SPW
    spw_and_cal_tables = []
    channel_ranges = get_channel_ranges_given_nspws_for_ms(
        ms=cal_ms, nspw=gain_cal_options.nspw
    )
    for idx, channel_range in enumerate(channel_ranges):
        logger.info(f"Calibrating {idx + 1} of {len(channel_ranges)}, {channel_range=}")
        spw_str = f"0:{channel_range[0]}~{channel_range[1]}"
        cal_table = create_and_check_caltable_path(
            ms=cal_ms, channel_range=channel_range
        )

        gaincal(
            container=casa_container,
            bind_dirs=(cal_ms.path.parent, cal_table.parent),
            vis=str(cal_ms.path),
            caltable=str(cal_table),
            spw=spw_str,
            solint=gain_cal_options.solint,
            gaintype=gain_cal_options.gaintype,
            minsnr=gain_cal_options.minsnr,
            calmode=gain_cal_options.calmode,
            selectdata=gain_cal_options.selectdata,
            uvrange=gain_cal_options.uvrange,
        )

        if not cal_table.exists():
            logger.critical(
                "The calibration table was not created. Likely gaincal failed. "
            )
            if raise_error_on_fail:
                raise GainCalError(f"Gaincal failed for {cal_ms.path}")
            else:
                return ms

        spw_and_cal_tables.append((spw_str, cal_table))

    # Now apply each of the solutions to the corresponding SPW.
    # Relying on the spw= channel selection to only be updating
    # the visibilities in the existing CORRECTED_DATA column,
    # not overwriting the entire column
    for idx, (spw_str, cal_table) in enumerate(spw_and_cal_tables):
        logger.info(
            f"Apply solutions {idx + 1} of {len(spw_and_cal_tables)}, {spw_str}"
        )
        applycal(
            container=casa_container,
            bind_dirs=(cal_ms.path.parent, cal_table.parent),
            vis=str(cal_ms.path),
            gaintable=str(cal_table),
            spw=spw_str,
            flagbackup=False,
        )

        if archive_cal_table:
            zip_folder(in_path=cal_table)

    flag_versions_table = cal_ms.path.with_suffix(".ms.flagversions")
    if flag_versions_table.exists():
        zip_folder(in_path=flag_versions_table)

    return cal_ms.with_options(column="CORRECTED_DATA")


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description=__doc__)

    sub_parsers = parser.add_subparsers(dest="mode")

    gaincal_parser = sub_parsers.add_parser("gaincal")
    gaincal_parser.add_argument(
        "ms", type=Path, help="Path to the measurement set to calibrate. "
    )
    gaincal_parser.add_argument(
        "--casa-container",
        type=Path,
        default=None,
        help="Path to the CASA6 singularity container",
    )
    gaincal_parser.add_argument(
        "--round", type=int, default=1, help="Self-calibration round number. "
    )
    gaincal_parser.add_argument(
        "--column", type=str, default="DATA", help="The column to self-calibrate"
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    if args.mode == "gaincal":
        gaincal_applycal_ms(
            ms=MS(path=args.ms, column=args.column),
            round=args.round,
            casa_container=args.casa_container,
        )


if __name__ == "__main__":
    cli()



"""Tests around utility helper functions for self-calibration"""

from __future__ import annotations

import shutil
from pathlib import Path

import pytest

from flint.options import MS
from flint.selfcal.utils import (
    consider_skip_selfcal_on_round,
    create_and_check_caltable_path,
    get_channel_ranges_given_nspws,
    get_channel_ranges_given_nspws_for_ms,
)
from flint.utils import get_packaged_resource_path


def test_create_solution_path():
    """Create the path to a solutions file, including appropriate handling of a channel range"""

    ms_path = Path("/jack/has/a/measurement_set.ms")
    cal_table = create_and_check_caltable_path(ms=MS.cast(ms_path))

    assert isinstance(cal_table, Path)
    assert cal_table == Path("/jack/has/a/measurement_set.caltable")

    cal_table = create_and_check_caltable_path(
        ms=MS.cast(ms_path), channel_range=(0, 143)
    )

    assert isinstance(cal_table, Path)
    assert cal_table == Path("/jack/has/a/measurement_set.caltable.ch0000-0143")


@pytest.fixture
def casda_example(tmpdir):
    ms_zip = Path(
        get_packaged_resource_path(
            package="flint.data.tests",
            filename="scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms.zip",
        )
    )
    outpath = Path(tmpdir) / "extract"

    shutil.unpack_archive(ms_zip, outpath)

    ms_path = (
        Path(outpath)
        / "scienceData.EMU_0529-60.SB50538.EMU_0529-60.beam08_averaged_cal.leakage.ms"
    )

    return ms_path


def test_get_channel_ranges_given_nspws_for_ms(casda_example):
    """Obtain channel ranges given an MS"""
    ms_path = Path(casda_example)
    chan_ranges = get_channel_ranges_given_nspws_for_ms(ms=ms_path, nspw=1)
    assert isinstance(chan_ranges, tuple)
    assert len(chan_ranges) == 1
    assert chan_ranges[0][0] == 0
    assert chan_ranges[0][1] == 287

    chan_ranges = get_channel_ranges_given_nspws_for_ms(ms=ms_path, nspw=4)
    assert isinstance(chan_ranges, tuple)
    assert len(chan_ranges) == 4
    assert chan_ranges[0][0] == 0
    assert chan_ranges[0][1] == 71

    assert chan_ranges[1][0] == 72
    assert chan_ranges[1][1] == 143
    assert chan_ranges[-1][1] == 287


def test_get_channel_ranges_given_nspws_for_ms_with_oddstep(casda_example):
    """Obtain channel ranges given an MS"""
    ms_path = Path(casda_example)
    chan_ranges = get_channel_ranges_given_nspws_for_ms(ms=ms_path, nspw=5)
    assert isinstance(chan_ranges, tuple)
    assert len(chan_ranges) == 5
    assert chan_ranges[0][0] == 0
    assert chan_ranges[0][1] == 57

    assert chan_ranges[1][0] == 58
    assert chan_ranges[1][1] == 115

    assert chan_ranges[-1][1] == 287


def test_get_channel_ranges_given_nspw():
    """At times channel ranges given a number of channels and number of windows needs to be created"""
    chan_ranges = get_channel_ranges_given_nspws(num_channels=288, nspws=1)
    assert isinstance(chan_ranges, tuple)
    assert len(chan_ranges) == 1
    assert chan_ranges[0][0] == 0
    assert chan_ranges[0][1] == 287

    chan_ranges = get_channel_ranges_given_nspws(num_channels=288, nspws=4)
    assert isinstance(chan_ranges, tuple)
    assert len(chan_ranges) == 4
    assert chan_ranges[0][0] == 0
    assert chan_ranges[0][1] == 71

    assert chan_ranges[1][0] == 72
    assert chan_ranges[1][1] == 143
    assert chan_ranges[-1][1] == 287


def test_get_channel_ranges_given_nspw_odd_steps():
    """At times channel ranges given a number of channels and number of windows needs to be created.
    This has odd channel steps"""
    chan_ranges = get_channel_ranges_given_nspws(num_channels=288, nspws=5)
    assert isinstance(chan_ranges, tuple)
    assert len(chan_ranges) == 5
    assert chan_ranges[0][0] == 0
    assert chan_ranges[0][1] == 57

    assert chan_ranges[1][0] == 58
    assert chan_ranges[1][1] == 115

    assert chan_ranges[-1][1] == 287


def test_consider_skip_selfcal():
    """Ensure that the skipping behaves as expected"""

    # None here means nothing should be skipped
    res = consider_skip_selfcal_on_round(current_round=1, skip_selfcal_on_rounds=None)

    assert not res

    res = consider_skip_selfcal_on_round(current_round=1, skip_selfcal_on_rounds=[2, 3])
    assert not res

    res = consider_skip_selfcal_on_round(current_round=2, skip_selfcal_on_rounds=[2, 3])
    assert res
    res = consider_skip_selfcal_on_round(current_round=2, skip_selfcal_on_rounds=2)
    assert res



"""Utilities related to running commands in a singularity container"""

from __future__ import annotations

from pathlib import Path
from subprocess import CalledProcessError
from time import sleep
from typing import Callable, Collection

from spython.main import Client as sclient

from flint.exceptions import AttemptRerunException
from flint.logging import logger
from flint.utils import get_job_info, log_job_environment


def pull_container(container_directory: Path, uri: str, file_name: str) -> Path:
    """Download a singularity container from an appropriate ``uri``.

    Args:
        container_directory (Path): Where to store the output container
        uri (str): The reference to the container to pull, e.g. docker://profile/container:tag
        file_name (str): The output name of the container

    Returns:
        Path: Location of the new container
    """
    logger.info(f"Attempting to pull {uri=} into {container_directory=}")

    container_path, output = sclient.pull(
        image=uri, pull_folder=str(container_directory), name=file_name, stream=True
    )
    for line in output:
        # Remove the trailing new line character
        logger.info(line.rstrip())

    container_path = Path(container_path)

    assert container_path.exists(), f"{container_path=} does not exist, but should"

    return container_path


def run_singularity_command(
    image: Path,
    command: str,
    bind_dirs: Path | Collection[Path] | None = None,
    stream_callback_func: Callable | None = None,
    ignore_logging_output: bool = False,
    max_retries: int = 2,
) -> None:
    """Executes a command within the context of a nominated singularity
    container

    Args:
        image (Path): The singularity container image to use
        command (str): The command to execute
        bind_dirs (Optional[Union[Path,Collection[Path]]], optional): Specifies a Path, or list of Paths, to bind to in the container. Defaults to None.
        stream_callback_func (Optional[Callable], optional): Provide a function that is applied to each line of output text when singularity is running and `stream=True`. IF provide it should accept a single (string) parameter. If None, nothing happens. Defaultds to None.
        ignore_logging_output (bool, optional): If `True` output from the executed singularity command is not logged. Defaults to False.
        max_reties (int, optional): If a callback handler is specified which raised an `AttemptRerunException`, this signifies how many attempts should be made. Defaults to 2.

    Raises:
        FileNotFoundError: Thrown when container image not found
        CalledProcessError: Thrown when the command into the container was not successful
    """
    if max_retries <= 0:
        raise ValueError("Too many retries")

    if not image.exists():
        raise FileNotFoundError(f"The singularity container {image} was not found. ")

    logger.info(f"Running {command} in {image}")

    job_info = log_job_environment()
    bind: None | list[str] = None
    if bind_dirs:
        logger.info("Preparing bind directories")
        if isinstance(bind_dirs, Path):
            bind_dirs = [bind_dirs]

        # Get only unique paths to avoid duplication in bindstring.
        # bind_str = ",".join(set([str(p.absolute().resolve()) for p in bind_dirs]))
        bind = (
            list(set([str(Path(p).absolute().resolve()) for p in bind_dirs]))
            if len(bind_dirs) > 0
            else None
        )

        logger.info(f"Constructed singularity bindings: {bind}")

    try:
        output = sclient.execute(
            image=image.resolve(strict=True).as_posix(),
            command=command.split(),
            bind=bind,
            return_result=True,
            quiet=False,
            stream=True,
            stream_type="both",
        )

        for line in output:
            if not ignore_logging_output:
                logger.info(line.rstrip())
            if stream_callback_func:
                stream_callback_func(line)

        # Sleep for a few moments. If the command created files (often they do), give the lustre a moment
        # to properly register them. You dirty sea dog.
        sleep(2.0)
    except AttemptRerunException as e:
        logger.info("A callback handler has raised an error. Attempting to rerun.")
        logger.info(f"{e=}")
        run_singularity_command(
            image=image,
            command=command,
            bind_dirs=bind_dirs,
            stream_callback_func=stream_callback_func,
            ignore_logging_output=ignore_logging_output,
            max_retries=max_retries - 1,
        )

    except CalledProcessError as e:
        logger.error(f"Failed to run command: {command}")
        logger.error(f"Stdout: {e.stdout}")
        logger.error(f"Stderr: {e.stderr}")
        logger.error(f"{e=}")
        job_info = get_job_info()
        if job_info:
            logger.error(f"{get_job_info()}")

        raise e


def singularity_wrapper(
    fn: Callable,
) -> Callable:
    """A decorator function to around another function that when executed
    returns a command to execute within a container. The returned function has
    the arguments as ``run_singularity_command``, and any unused keywords are
    passed to the wrapped function.

    Args:
        fn (Callable): The function that generates the command to execute

    Returns:
        Callable: Wrapper function
    """

    def wrapper(
        container: Path,
        bind_dirs: Path | Collection[Path] | None = None,
        stream_callback_func: Callable | None = None,
        ignore_logging_output: bool = False,
        **kwargs,
    ) -> str:
        """Function that can be used as a decorator on an input function. This function
        should generate and return a command that will be executed within the specified container.

        Args:
            fn (Callable): The function that generates a command to execute. All ``**kwargs`` are passed to this function
            container (Path): Path to the container that will be usede to execute the generated command
            bind_dirs (Optional[Union[Path,Collection[Path]]], optional): Specifies a Path, or list of Paths, to bind to in the container. Defaults to None.
            stream_callback_func (Optional[Callable], optional): Provide a function that is applied to each line of output text when singularity is running and `stream=True`. IF provide it should accept a single (string) parameter. If None, nothing happens. Defaultds to None.
            ignore_logging_output (bool, optional): If `True` output from the executed singularity command is not logged. Defaults to False.

        Returns:
            str: The command that was executed
        """

        task_str = fn(**kwargs)

        assert isinstance(task_str, str), f"{task_str=}, but needs to be a string"

        logger.info(f"wrapper {task_str=}")
        logger.info(f"wrapper {bind_dirs=}")

        run_singularity_command(
            image=container,
            command=f"{task_str}",
            bind_dirs=bind_dirs,
            ignore_logging_output=ignore_logging_output,
            stream_callback_func=stream_callback_func,
        )

        return task_str

    return wrapper



"""Construct a leakge map between two polarisations, typically V/I"""

from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import NamedTuple, Union

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS
from astropy.wcs.utils import skycoord_to_pixel

from flint.catalogue import guess_column_in_table
from flint.logging import logger

TableOrPath = Union[Table, Path]


class LeakageFilters(NamedTuple):
    """Description of the filtering options to apply to components
    when characterising leakage"""

    isolation_radius_deg: float = 0.0155
    """The minimum distance to the nearest component"""
    upper_int_peak_ratio: float = 1.2
    """The upper limit on acceptable int/peak ratios"""
    lower_int_peak_ratio: float = 0.8
    """The lower limit on acceptable int/peak ratios"""
    search_box_size: int = 1
    """The size of a box to search for peak polarised signal in"""
    noise_box_size: int = 30
    """the size of a box to compute a local RMS noise measure from"""
    mean_box_size: int = 10
    """The size of abox to compute a local mean measure over"""
    source_snr: float = 40
    """Minimum stokes-I signal-to-noise ratio"""


class FITSImage(NamedTuple):
    """Container to couple FITS header, image and WCS"""

    data: np.ndarray
    """The data of the fits image"""
    header: dict
    """Header of the fits image"""
    wcs: WCS
    """Celestial WCS of the fits image"""
    path: Path
    """Path of the loaded FITS image on disk"""


class PixelCoords(NamedTuple):
    """Slim container to help collect and maintain pixel coordinates. Not intended for extensive use"""

    y: np.ndarray
    """The y-coordinate of a set of pixels"""
    x: np.ndarray
    """The x-coordinate of a set of pixels"""


def _load_fits_image(fits_path: Path) -> FITSImage:
    """Load in a FITS image and package the components into a consistent
    form. Not intended for extensive use.

    Args:
        fits_path (Path): The path of the FITS image to examining

    Returns:
        FITSImage: Loaded FITS properties
    """

    assert fits_path.suffix == ".fits", (
        f"Unexpected file type for {fits_path=}, expected fits"
    )
    logger.info(f"Opening {fits_path=}")
    with fits.open(fits_path) as in_fits:
        image_data = in_fits[0].data  # type: ignore
        header = dict(in_fits[0].header.items())  # type: ignore
        wcs = WCS(header)

    return FITSImage(data=image_data, header=header, wcs=wcs, path=fits_path)


def _load_component_table(catalogue: TableOrPath) -> Table:
    """Return a table given either a loaded table or a path to a table on disk"""
    component_table = (
        Table.read(catalogue) if isinstance(catalogue, Path) else catalogue
    )
    logger.info(f"Loaded component table of {len(component_table)}")

    return component_table


def filter_components(
    table: Table,
    peak_col: str,
    int_col: str,
    int_err_col: str,
    leakage_filters: LeakageFilters,
    ra_col: str | None = None,
    dec_col: str | None = None,
) -> Table:
    """Apply the pre-processing operations to catalogue components to select an
    optimal sample of sources for leakage characterisation. Sources will be selected
    based on:

    * how isolated they are
    * compactness, as traced by their int/peak

    Args:
        table (Table): Collection of sources, as produced from a source finder
        peak_col (str): The column name describing the peak flux density
        int_col (str): The column name describing integrated flux
        int_err_col (str): The column container errors that correspond to `int_col` to use when computing signal-to-noise
        leakage_filters (LeakageFilters): Criteria applied to the source components in the table
        ra_col (Optional[str], optional): The RA column name. If None, it will be guessed. Defaults to None.
        dec_col (Optional[str], optional): The Dec column name. If None, it will be guessed. Defaults to None.

    Returns:
        Table: A filtered table
    """
    ra_col = guess_column_in_table(table=table, column="ra", guess_column=ra_col)
    dec_col = guess_column_in_table(table=table, column="dec", guess_column=dec_col)

    assert all(
        [
            col in table.colnames
            for col in (ra_col, dec_col, peak_col, int_col, int_err_col)
        ]
    ), (
        f"Supplied column names {ra_col=} {dec_col=} {peak_col=} {int_col=} partly missing from {table.colnames=}"
    )

    total_comps = len(table)
    sky_coords = SkyCoord(table[ra_col], table[dec_col], unit=(u.deg, u.deg))

    # The match_to_catalog_sky return idx, sep2d, sep3d. We care about separation, matey
    isolation_mask = sky_coords.match_to_catalog_sky(sky_coords, nthneighbor=2)[1] > (
        leakage_filters.isolation_radius_deg * u.deg
    )  # type: ignore
    logger.info(
        f"{np.sum(isolation_mask)} of {total_comps} sources are isolated with radius {float(leakage_filters.isolation_radius_deg):.5f} deg"
    )

    ratio = table[int_col] / table[peak_col]  # type: ignore
    ratio_mask = (leakage_filters.lower_int_peak_ratio < ratio) & (
        ratio < leakage_filters.upper_int_peak_ratio
    )  # type: ignore
    logger.info(
        f"{np.sum(ratio_mask)} of {total_comps} sources are compact with {leakage_filters.lower_int_peak_ratio} < int/peak < {leakage_filters.upper_int_peak_ratio}"
    )

    signal_to_noise = table[int_col] / table[int_err_col]  # type: ignore
    signal_mask = signal_to_noise > leakage_filters.source_snr
    logger.info(
        f"{np.sum(signal_mask)} of {total_comps} sources have S/N above {leakage_filters.source_snr}"
    )

    mask = isolation_mask & ratio_mask & signal_mask
    logger.info(
        f"{np.sum(mask)} of {total_comps} sources are isolated, compact and bright"
    )

    table = table[mask]  # type: ignore

    return table


def get_xy_pixel_coords(
    table: Table,
    wcs: WCS,
    ra_col: str | None = None,
    dec_col: str | None = None,
) -> PixelCoords:
    """Convert (RA, Dec) positions in a catalogue into (x, y)-pixels given an WCS

    Args:
        table (Table): The table containing sources to collect (x, y)-coordinates
        wcs (WCS): The WCS description to use to resolve (RA, Dec) to (x, y)
        ra_col (Optional[str], optional): The RA column name. If None, it will be guessed. Defaults to None.
        dec_col (Optional[str], optional): The Dec column name. If None, it will be guessed. Defaults to None.

    Returns:
        PixelCoords: _description_
    """
    ra_col = guess_column_in_table(table=table, column="ra")
    dec_col = guess_column_in_table(table=table, column="dec")

    sky_coord = SkyCoord(table[ra_col], table[dec_col], unit=(u.deg, u.deg))
    x, y = skycoord_to_pixel(wcs=wcs.celestial, coords=sky_coord, origin=0)

    return PixelCoords(y=np.floor(y).astype(int), x=np.floor(x).astype(int))


def load_and_filter_components(
    catalogue: TableOrPath, leakage_filters: LeakageFilters
) -> Table:
    """Load in a component catalogue table and apply filters to them. The
    remaining components will be used to characterise leakage

    Args:
        catalogue (TableOrPath): The path to a component catalogue, or a loaded component catalogue
        leakage_filters (LeakageFilters): Filtering options to find ideal components for leakage characterisation

    Returns:
        Table: Filtered component catalogue
    """
    # TODO: Need to have a single guess function here
    comp_table = _load_component_table(catalogue=catalogue)
    ra_col = guess_column_in_table(table=comp_table, column="ra")
    dec_col = guess_column_in_table(table=comp_table, column="dec")
    peak_col = guess_column_in_table(table=comp_table, column="peakflux")
    int_col = guess_column_in_table(table=comp_table, column="intflux")
    int_err_col = guess_column_in_table(table=comp_table, column="intfluxerr")

    # TODO: Need to use Catalogue here
    comp_table = filter_components(
        table=comp_table,
        ra_col=ra_col,
        dec_col=dec_col,
        peak_col=peak_col,
        int_col=int_col,
        int_err_col=int_err_col,
        leakage_filters=leakage_filters,
    )
    return comp_table


class PolStatistics(NamedTuple):
    """Simple container for statistics around the extraction of leakage polarisation statistics"""

    peak: np.ndarray
    """The peak pixel value"""
    noise: np.ndarray
    """The standard deviation of pixels within a box"""
    mean: np.ndarray
    """The mean of pixels within a box"""


def extract_pol_stats_in_box(
    pol_image: np.ndarray,
    pixel_coords: PixelCoords,
    search_box_size: int,
    noise_box_size: int,
    mean_box_size: int,
) -> PolStatistics:
    """Construct two boxes around nominated pixel coordinates to:

    * extract the peak signal within
    * calculate a local RMS value for

    Args:
        pol_image (np.ndarray): The loaded polarised image
        pixel_coords (PixelCoords): Collection of pixel positioncs to evaluate the peak polarisation and noise at
        search_box_size (int): Size of box to extract the maximum polarised signal from
        noise_box_size (int): Size of box to calculate the RMS over
        mean_box_size (int): Size of box to calculate an mean over

    Returns:
       PolStatistics: Extracted statistics, including peak polarised signal, noise and mean
    """
    y_max, x_max = pol_image.shape[-2:]

    logger.info(f"{pol_image.shape=}, extracted {y_max=} and {x_max=}")

    pol_peak = None
    pol_noise = None
    pol_mean = None
    # TODO: This loop should be reformed to allow iterating over something that defines the mode and box
    for idx, box_size in enumerate((search_box_size, noise_box_size, mean_box_size)):
        box_delta = int(np.ceil(box_size / 2))

        y_edge_min = np.maximum(pixel_coords.y - box_delta, 0).astype(int)
        y_edge_max = np.minimum(pixel_coords.y + box_delta, y_max - 1).astype(int)

        x_edge_min = np.maximum(pixel_coords.x - box_delta, 0).astype(int)
        x_edge_max = np.minimum(pixel_coords.x + box_delta, x_max - 1).astype(int)

        assert not np.any(y_edge_min == y_edge_max), "The y box edges are equal"
        assert not np.any(x_edge_min == x_edge_max), "The x box edges are equal"

        pol_image = np.squeeze(pol_image)

        # search_box = np.squeeze(pol_image)[y_edge_min:y_edge_max, x_edge_min:x_edge_max]
        search_box = [
            pol_image[y_min:y_max, x_min:x_max].flatten()
            for (y_min, y_max, x_min, x_max) in zip(
                y_edge_min, y_edge_max, x_edge_min, x_edge_max
            )
        ]

        if idx == 0:
            pol_peak = np.array(
                [
                    (
                        data[np.nanargmax(np.abs(data))]
                        if np.any(np.isfinite(data))
                        else np.nan
                    )
                    for data in search_box
                ]
            )
        elif idx == 1:
            pol_noise = np.array(
                [
                    np.nanstd(data) if np.any(np.isfinite(data)) else np.nan
                    for data in search_box
                ]
            )
        elif idx == 2:
            pol_mean = np.array(
                [
                    np.nanmean(data) if np.any(np.isfinite(data)) else np.nan
                    for data in search_box
                ]
            )

    assert pol_peak is not None, f"{pol_peak=}, which should not happen"
    assert pol_noise is not None, f"{pol_noise=}, which should not happen"
    assert pol_mean is not None, f"{pol_mean=}, which should not happen"

    pol_stats = PolStatistics(peak=pol_peak, noise=pol_noise, mean=pol_mean)

    return pol_stats


def _get_output_catalogue_path(
    input_path: Path, pol: str, output_path: Path | None = None
) -> Path:
    """Create the output leakage catalogue name"""
    # NOTE: This is a separate function to test against after a silly. Might move with the other named Pirates
    assert isinstance(input_path, Path)
    input_suffix = input_path.suffix

    output_path = (
        input_path.with_suffix(f".{pol}_leakage{input_suffix}")
        if output_path is None
        else output_path
    )
    assert output_path is not None, (
        f"{output_path=} is empty, and no catalogue path provided"
    )

    return Path(output_path)


def create_leakge_component_table(
    pol_image: Path,
    catalogue: Table | Path,
    pol: str = "v",
    output_path: Path | None = None,
) -> Path:
    """Create a component catalogue that includes enough information to describe the
    polarisation fraction of sources across a field. This is intended to be used
    for leakage characterisation.

    New catalogue columns will be added:

    * pol_fraction: The POL/I fraction. The peak flux is taken from the catalogue, using the appropriate column name
    * pol_peak: The peak polarised signal in the nearby region of a component position
    * pol_noise: The noise in the polarised image pixels around the component position

    Args:
        pol_image (Path): The polarised image that will be used to extract peak polarised flux from
        catalogue (Union[Table, Path]): Component table describing positions to extract flux from
        pol (str, optional): The polarisation stokes being considered. Defaults to "v".
        output_path (Optional[Path], optional): The path of the new catalogue. If `None` it is derived from the input `catalogue` path. Defaults to None.

    Returns:
        Path: Path to the new catalogue use for leakage
    """
    pol_fits = _load_fits_image(fits_path=pol_image)

    leakage_filters = LeakageFilters()

    components = load_and_filter_components(
        catalogue=catalogue, leakage_filters=leakage_filters
    )
    peak_flux_col = guess_column_in_table(table=components, column="peakflux")
    i_values = components[peak_flux_col]

    pol_pixel_coords = get_xy_pixel_coords(table=components, wcs=pol_fits.wcs)
    pol_stats = extract_pol_stats_in_box(
        pol_image=pol_fits.data,
        pixel_coords=pol_pixel_coords,
        search_box_size=leakage_filters.search_box_size,
        noise_box_size=leakage_filters.noise_box_size,
        mean_box_size=leakage_filters.mean_box_size,
    )
    frac_values = pol_stats.peak / i_values

    logger.info(f"{frac_values.shape=}")
    components[f"{pol}_fraction"] = frac_values
    components[f"{pol}_peak"] = pol_stats.peak
    components[f"{pol}_noise"] = pol_stats.noise
    components[f"{pol}_mean"] = pol_stats.mean

    output_path = _get_output_catalogue_path(
        input_path=catalogue if isinstance(catalogue, Path) else pol_image,
        pol=pol,
        output_path=output_path,
    )

    logger.info(f"Writing {output_path}")
    components.write(output_path, overwrite=True)

    return output_path


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(description="Create a leakage catalogue and map")
    parser.add_argument("pol_image", type=Path, help="Path to the polarisation image")
    parser.add_argument(
        "component_catalogue", type=Path, help="Path to the component catalogue"
    )
    parser.add_argument(
        "--pol",
        default="v",
        choices=("q", "u", "v"),
        help="The polarisation that is being considered",
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    create_leakge_component_table(
        pol_image=args.pol_image,
        catalogue=args.component_catalogue,
        pol=args.pol,
    )


if __name__ == "__main__":
    cli()
